"""
Script to edit GFF files

Print Attributes in a GFF file
******************************

By default reads the first 10 lines of a GFF file and prints all attributes
present in the file sorted. Not all annotations may have the same set of
attributes, so a higher number of lines may be necessary to be read.

View GFF
********

Used to print the content of a GFF file as a table (more output formats will be
added later).

The attributes printed are passed with `-a`, one attribute at a time. For
example::

    edit-gff view -a uid test.gff

will print `uid` for all annotations. Multiple attributes can be passed, like::

    edit-gff view -a uid -a taxon_id test.gff

that will print a table with `uid` and `taxon_id` of each annotation.

The default behaviour is to print only annotations that have all the attributes
requested. This can be changed by using the `-k` options and the fields that
were not found are empty strings.

An header can be printed with the `-h` option.

.. note::

    the order of the fields in the table is the same as the order of the
    attributes passed with `-a`

Change or Add Attributes
************************

Add or changes annotations in a GFF files with the specified attributes.

The attributes and the values are passed with the `-a` option, for example
to set all annotations `taxon_id` to 2, you can pass `-a taxon_id 2`. Multiple
attributes can be set by passing multiple options. For example::

    edit-gff add -a taxon_id 2 -a taxon_db CUSTOM test.gff

will set the `taxon_id` to 2 and the `taxon_db` to CUSTOM for all annotations.

The default behaviour is to not change an attribute already set in an
annotation, but this can be changed by passing the `-w` option. Moreover,
only edited annotations can be output with `-o`.

To change attributes on a subset of the annotations, a file can be passed with
the `-f` options, which contains one `uid` per line. Only annotations that
match a `uid` in that list are edited.

Remove Attributes
*****************

Removes a list of attributes in a GFF file. Only attributes in the last column
of a GFF file (fields separated by a ';') can be removed. Attributes are passed
with the `-a` option followed by one attribute. Multiple `-a attribute` options
can be passed.

To remove attributes on a subset of the annotations, a file can be passed with
the `-f` options, which contains one `uid` per line. Only annotations that
match a `uid` in that list are edited.

Table
*****

Similar to the *add* command and with similar functions as *add-gff-info
addtaxa*, it allows the adding/changing of attributes from a table file.

The user defines 2 attributes in a GFF annotation, the *key* and the
*attribute*. The *key* is used to find if an annotation is to be modified and
the *attribute* is set for that annotation with the value in the table. For
example a table::

    GENE001,1.1.3.3
    GENE002,1.2.3.3

If *key* chosen is *gene_id* and *attribute* is *EC*, the GFF will be scanned
for annotation that have the *gene_id* equal to *GENE001* and set the
*attribute* *EC* to *1.1.3.3* and similarly for the second row.

The table can have multiple fields, but only 2 can be loaded, the *key* and
*attribute* in the options. The 2 fields are loaded into a python dictionary,
with the key and attribute being respectively the key and value in it. So 2
things must be noted:

    1) duplicates keys will be overwritten (only last one remains)
    2) the entire fields are first loaded, which can take up a lot of RAM

The default is for the key to be the first field (0) and the attribute is the
second (1). The table may contains some headers, so the first N rows can be
skipped with `-r`. Also, the field separator can be chosen, as well as only
the edited annotation be printed (`-o` option).

Changes
*******

.. versionadded:: 0.4.4

"""

import logging
import click
from . import utils
from .. import logger
from mgkit.io import gff, open_file
from mgkit.utils.dictionary import text_to_dict

LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


@main.command('fields', help="""Prints the fields in a GFF File""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-n', '--num-ann', type=click.INT, show_default=True,
              default=10, help='''Number of annotations to parse, 0 will parse
              the whole file''')
@click.argument('gff_file', type=click.File('rb', lazy=False), default='-')
@click.argument('txt_file', type=click.File('w', lazy=False), default='-')
def view_fields(verbose, num_ann, gff_file, txt_file):

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    fields = set()

    for count, annotation in enumerate(gff.parse_gff(gff_file)):
        fields.update(annotation.to_dict())
        if num_ann > 0:
            if (count + 1) == num_ann:
                break

    txt_file.write('\n'.join(sorted(fields)) + '\n')


@main.command('add', help="""Add fields to a GFF File""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-a', '--attributes', multiple=True, nargs=2, required=True,
              help="""Add attributes to the GFF file. For example `-a taxon_id 2` will add taxon_id attribute with a value of 2 to all annotations. Multiple attributes can be set, for example: `-a taxon_id 2 -a gene_id TEST`""")
@click.option('-w', '--overwrite', is_flag=True,
              help="Overwrite the attributes if present")
@click.option('-o', '--only-edited', is_flag=True,
              help="Only output edited annotations")
@click.option('-f', '--uids', type=click.File('r', lazy=False), default=None,
              help="Only edit annotations with `uid` in a file (one per line)")
@click.argument('input_file', type=click.File('rb', lazy=False), default='-')
@click.argument('output_file', type=click.File('wb', lazy=False), default='-')
def add_fields(verbose, attributes, overwrite, only_edited, uids, input_file,
               output_file):

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    if overwrite:
        LOG.info("Attributes/Values will be overwritten")

    LOG.info("Fields to add/change: %s", ', '.join(x[0] for x in attributes))

    if len(set((x[0] for x in attributes))) != len(attributes):
        utils.exit_script("Found duplicates attributes to edit", 1)

    if uids is not None:
        uids = set(line.strip() for line in uids)
        LOG.info("Number of `uid` passed: %d", len(uids))

    for annotation in gff.parse_gff(input_file):
        if (uids is not None) and (annotation.uid not in uids):
            if not only_edited:
                annotation.to_file(output_file)
            continue

        change_attr = set()
        for attribute, value in attributes:
            if (not overwrite) and annotation.has_attr(attribute):
                continue

            annotation.set_attr(attribute, value)
            change_attr.add(attribute)

        if change_attr or (not only_edited):
            annotation.to_file(output_file)


@main.command('remove', help="""Remove fields from a GFF File""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-a', '--attributes', multiple=True, required=True,
              help="""Remove attributes to the GFF file. For example `-a taxon_id` will remove taxon_id attribute for all annotations. Multiple attributes can be removed, for example: `-a taxon_id -a gene_id`""")
@click.option('-f', '--uids', type=click.File('r', lazy=False), default=None,
              help="Only edit annotations with `uid` in a file (one per line)")
@click.argument('input_file', type=click.File('rb', lazy=False), default='-')
@click.argument('output_file', type=click.File('wb', lazy=False), default='-')
def remove_fields(verbose, attributes, uids, input_file, output_file):

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    attributes = set(x.strip() for x in attributes)

    LOG.info("Fields to remove: %s", ', '.join(attributes))

    if uids is not None:
        uids = set(line.strip() for line in uids)
        LOG.info("Number of `uid` passed: %d", len(uids))

    for annotation in gff.parse_gff(input_file):
        if (uids is not None) and (annotation.uid not in uids):
            annotation.to_file(output_file)
            continue

        for attribute in attributes:
            annotation.del_attr(attribute)

        annotation.to_file(output_file)


@main.command('view', help="""View GFF file as table/json, etc.""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-h', '--header', is_flag=True, help='Print Header')
@click.option('-k', '--keep-empty', is_flag=True,
              help='Keep annotations where not all attributes were found')
@click.option('-a', '--attributes', multiple=True, required=True,
              help="""Attributes of GFF file to print. For example `-a taxon_id` will print `taxon_id` for all annotations. Multiple attributes can be printed, for example: `-a taxon_id -a gene_id`""")
@click.option('-s', '--separator', default='\t',
              help="Fields separator, default to `TAB`")
@click.argument('input_file', type=click.File('rb', lazy=False), default='-')
@click.argument('output_file', type=click.File('w', lazy=False), default='-')
def print_fields(verbose, header, keep_empty, attributes, separator,
                 input_file, output_file):

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    attributes = list(x.strip() for x in attributes)
    LOG.info("Fields to print: %s", ', '.join(attributes))
    if header:
        output_file.write('{}\n'.format(separator.join(attributes)))

    for annotation in gff.parse_gff(input_file):
        values = []
        for attribute in attributes:
            try:
                values.append(annotation.get_attr(attribute))
            except gff.AttributeNotFound:
                if not keep_empty:
                    continue
                values.append('')

        if len(attributes) == len(values):
            output_file.write('{}\n'.format(separator.join(values)))


@main.command('table', help="""Adds fields from a Table file""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-k', '--key', default='uid',
              help="Attribute used to search the table defaults to `uid`")
@click.option('-a', '--attribute', required=True,
              help="Attribute to add/change")
@click.option('-o', '--only-edited', is_flag=True,
              help="Only output edited annotations")
@click.option('-r', '--skip-rows', default=0,
              help="Number of rows to skip at the start of the file")
@click.option('-s', '--separator', default='\t',
              help="Fields separator, default to `TAB`")
@click.option('-t', '--table-file', type=click.File('rb', lazy=False),
              required=True)
@click.option('-ki', '--key-index', default=0, show_default=True,
              help="Which field in the table is the key value")
@click.option('-ai', '--attr-index', default=1, show_default=True,
              help="Which field in the table is the attribute value")
@click.argument('input_file', type=click.File('rb', lazy=False), default='-')
@click.argument('output_file', type=click.File('wb', lazy=False), default='-')
def add_fields_from_table(verbose, key, attribute, only_edited, skip_rows,
                          separator, table_file, key_index, attr_index,
                          input_file, output_file):

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info("Key used is '%s' and attribute '%s'", key, attribute)
    LOG.info("N. rows skipped '%d' Key index is '%d' and attribute index '%d'",
             skip_rows, key_index, attr_index)

    if getattr(table_file, 'name', None) is not None:
        LOG.info("Reading values from (%s)", table_file.name)

    fields = dict(
        text_to_dict(open_file(table_file), skip_lines=skip_rows,
                     sep=separator, key_index=key_index,
                     value_index=attr_index, encoding='ascii')
    )

    changed = 0

    for annotation in gff.parse_gff(input_file):
        try:
            key_ann_value = annotation.get_attr(key)
        except gff.AttributeNotFound:
            if only_edited:
                continue
        try:
            annotation.set_attr(attribute, fields[key_ann_value])
            changed += 1
        except KeyError:
            if only_edited:
                continue
        annotation.to_file(output_file)

    LOG.info('Changed %d annotations', changed)



"""
Module containing functions related to sequence data

.. note::

    For those functions without a docstring, look at the same with a
    underscore ('_') prepended.

"""
from __future__ import division  # add check to use only on python 2.x
from builtins import range, zip
from future.utils import viewitems, viewvalues
import itertools
import logging
import random
import collections
import numpy
import pandas
from scipy import stats
from scipy import interpolate
import statsmodels.api as sm

from ..utils.common import between
from .trans_tables import UNIVERSAL
from ..io import fasta
from ._sequence import get_kmers, sliding_window, sequence_signature, \
    signatures_matrix

LOG = logging.getLogger(__name__)

TRANS_TABLE = UNIVERSAL
"""Translation table - Universal genetic code"""

REV_COMP = {
    'A': 'T',
    'T': 'A',
    'C': 'G',
    'G': 'C'
}
"Dictionary containing the complement of each nucleotide sequence"


def make_reverse_table(tbl=None):
    """
    Makes table to reverse complement a sequence by :func:`reverse_complement`.
    The table used is the complement for each nucleotide, defaulting to
    :data:`REV_COMP`
    """
    if tbl is None:
        tbl = REV_COMP
    trans_table = [chr(idx) for idx in range(256)]
    for nuc, rev in viewitems(tbl):
        trans_table[ord(nuc)] = rev
    return ''.join(trans_table)


# The maketrans function is not available on Python3
REV_COMP_ASCII = u''.maketrans(REV_COMP)


def reverse_complement_old(seq, tbl=None):
    """
    Returns the reverse complement of a nucleotide sequence

    :param str seq: nucleotide sequence with uppercase characters
    :param dict tbl: dictionary of complement bases, like :data:`REV_COMP`

    :return str: returns the reverse complement of a nucleotide sequence
    """
    if tbl is None:
        tbl = REV_COMP
    rev_complement = [tbl[x] for x in seq[::-1]]
    return ''.join(rev_complement)


def reverse_complement(seq, tbl=REV_COMP_ASCII):
    """
    Returns the reverse complement of a nucleotide sequence

    :param str seq: nucleotide sequence with uppercase characters
    :param dict tbl: translation table returned by :func:`make_reverse_table`

    :return str: returns the reverse complement of a nucleotide sequence
    """
    return seq[::-1].translate(tbl)


def translate_sequence(sequence, start=0, tbl=None, reverse=False):
    """
    Translate a nucleotide sequence in an amino acid one.

    :param str sequence: sequence to translate, it's expected to be all caps
    :param int start: 0-based index for the translation to start
    :param dict tbl: dictionary with the translation for each codon
    :param bool reverse: if True, :func:`reverse_complement` will be called and
        the returned sequence translated

    :return str: the translated sequence
    """
    if reverse:
        sequence = reverse_complement(sequence)
    if tbl is None:
        tbl = UNIVERSAL

    trs = []

    for idx in range(start, len(sequence), 3):
        codon = sequence[idx:idx+3]
        if len(codon) < 3:
            break
        try:
            trs.append(tbl[codon])
        except KeyError:
            trs.append('X')
    return ''.join(trs)


def put_gaps_in_nuc_seq(nuc_seq, aa_seq, trim=True):
    """
    Match the gaps in an amino-acid aligned sequence to its original nucleotide
    sequence. If the nucleotide sequence is not a multiple of 3, the trim
    option by default trim those bases from the output.

    :param str nuc_seq: original nucleotide sequence
    :param str aa_seq: aligned amino-acid sequence
    :param bool trim: if True trim last nucleotide(s)

    :return str: gapped nucleotide sequence
    """
    # transform a string into a list of codons, then inserts the gaps
    nuc_alg = [nuc_seq[idx:idx+3] for idx in range(0, len(nuc_seq), 3)]

    for idx, aa_pos in enumerate(aa_seq):
        if aa_pos == '-':
            nuc_alg.insert(idx, '-' * 3)

    nuc_alg = ''.join(nuc_alg)

    if trim and (len(nuc_alg) != len(aa_seq) * 3):
        trim_idx = len(nuc_seq) % 3
        if trim_idx > 0:
            nuc_alg = nuc_alg[:-trim_idx]

    return nuc_alg


def get_syn_matrix(trans_table=None, nuc_list=None):
    """
    Returns a dictionary containing the expected count of synonymous and
    non-synonymous changes that a codon can have if one base is allowed to
    change at a time.

    There are 9 possible changes per codon.

    :param dict trans_table: a tranlation table, defaults to
        :data:`seq_utils.TRANS_TABLE`
    :param iterable nuc_list: a list of nucleotides in which a base can change,
        default to the keys of :data:`seq_utils.REV_COMP`

    :return dict: returns a dictionary in which for each codon a dictionary
        {'syn': 0, 'nonsyn': 0} holds the number of expected changes
    """

    if trans_table is None:
        trans_table = TRANS_TABLE
    if nuc_list is None:
        nuc_list = REV_COMP

    syn_matrix = dict(
        (codon, {'syn': 0, 'nonsyn': 0})
        for codon in trans_table
    )

    for codon1 in trans_table:
        for frame in range(0, 3):
            for nuc in nuc_list:
                codon2 = list(codon1)
                codon2[frame] = nuc
                codon2 = ''.join(codon2)

                # skips if its the same
                if codon1 == codon2:
                    continue

                if trans_table[codon1] == trans_table[codon2]:
                    syn_matrix[codon1]['syn'] += 1
                else:
                    syn_matrix[codon1]['nonsyn'] += 1

    return syn_matrix


def get_syn_matrix_all(trans_table=None):
    """
    Same as :func:`get_syn_matrix` but a codon can change in any of the ones
    included in trans_table.

    There are 63 possible changes per codon.
    """

    if trans_table is None:
        trans_table = TRANS_TABLE

    syn_matrix = dict(
        (codon, {'syn': 0, 'nonsyn': 0})
        for codon in trans_table
    )

    for codon1, codon2 in itertools.permutations(trans_table, 2):
        if trans_table[codon1] == trans_table[codon2]:
            syn_matrix[codon1]['syn'] += 1
        else:
            syn_matrix[codon1]['nonsyn'] += 1

    return syn_matrix


def get_seq_expected_syn_count(seq, start=0, syn_matrix=None):
    """
    .. versionchanged:: 0.4.4
        if codon contains *N* or is not a multiple of 3, it is skipped

    Calculate the expected number of synonymous and non-synonymous changes in a
    nucleotide sequence. Assumes that the sequence is already in the correct
    frame and its length is a multiple of 3.

    :param iterable seq: nucleotide sequence (uppercase chars)
    :param int start: frame of the sequence
    :param dict syn_matrix: dictionary that contains the expected number of
        changes for a codon, as returned by :func:`get_syn_matrix`

    :return tuple: tuple with counts of expected counts (syn, nonsyn)
    """
    if syn_matrix is None:
        syn_matrix = get_syn_matrix()

    seq_syn = 0
    seq_nonsyn = 0

    for idx in range(start, len(seq), 3):
        codon = seq[idx:idx+3]
        # if an ambiguity *N* is found, skip the codon
        if ('N' in codon) or (len(codon) != 3):
            continue
        seq_syn += syn_matrix[codon]['syn']
        seq_nonsyn += syn_matrix[codon]['nonsyn']

    return seq_syn, seq_nonsyn


def get_variant_sequence(seq, *snps):
    """
    .. versionadded:: 0.1.16

    Return a sequence changed in the positions requested.

    Arguments:
        seq (str): a sequence
        *snps (tuple): each argument passed is a tuple with the first element
            as a position in the sequence (1-based index) and the second
            element is the character to substitute in the sequence

    Returns:
        str: string with the changed characters

    Example:
        >>> get_variant_sequence('ACTGATATATGCGCGCATCT', (1, 'C'))
        'CCTGNTGTATGCGCGCATCT'

    .. note::

        It is used for nucleotide sequences, but it is valid to use any string
    """
    seq = list(seq)

    for pos, change in snps:
        seq[pos - 1] = change

    return ''.join(seq)


def convert_aa_to_nuc_coord(start, end, frame=0):
    """
    Converts aa coordinates to nucleotidic ones. The coordinates must be from
    '+' strand. For the '-' strand, use :func:`reverse_aa_coord` first.

    Arguments:
        start (int): start of the annotation (lowest number)
        end (int): end of the annotation (highest number)
        frame (int): frame of the AA translation (0, 1 or 2)

    Returns:
        tuple: the first element is the converted *start* and the second
        element is the converted *end*

    .. note::

        the coordinates are assumed to be 1-based indices

    """
    start = (start - 1) * 3 + 1  # gets the first base of the codon
    end = (end - 1) * 3 + 3  # gets the third base of the codon

    return (start + frame, end + frame)


def reverse_aa_coord(start, end, seq_len):
    """
    Used to reverse amino-acid coordinates when parsing an AA annotation on
    the - strand. Used when the BLAST or HMMER annotations use AA sequences.

    Arguments:
        start (int): start of the annotation
        end (int): end of the annotation
        seq_len (int): aa sequence length

    Returns:
        tuple: reversed (from strand - to strand +) coordinates. The first
        element is the converted *start* and the second element is the
        converted *end*

    .. note::

        * start and end are 1-based indices

    """

    return (seq_len - end + 1, seq_len - start + 1)


def calc_n50(seq_lengths):
    """
    Calculate the N50 statistics for a :class:`numpy.array` of sequence
    lengths.

    The algorithm finds in the supplied array the element (contig length) for
    which the sum all contig lengths equal or greater than it is equal to half
    of all assembled base pairs.

    :param array seq_lengths: an instance of a numpy array containing the
        sequence lengths

    :return int: the N50 statistics value
    """
    LOG.info("Calculating N50 for assembly")
    seq_lengths.sort()

    bp_half = seq_lengths.sum() / 2.0
    bp_sum = 0

    for seq_length in seq_lengths[::-1]:
        bp_sum += seq_length
        if bp_sum > bp_half:
            return seq_length


class Alignment(object):
    "Simple alignment class"
    _names = None
    _seqs = None

    def __init__(self, seqs=None):
        self._names = []
        self._seqs = []
        if seqs:
            self.add_seqs(seqs)

    def add_seq(self, name, seq):
        """
        Add a sequence to the alignment

        :param str name: name of the sequence
        :param str seq: sequence
        """
        self._names.append(name)
        self._seqs.append(seq)

    def add_seqs(self, seqs):
        """
        Add sequences to the alignment

        :param iterable seqs: iterable that returns (name, seq)
        """
        for name, seq in seqs:
            self.add_seq(name, seq)

    def get_position(self, pos):
        """
        Get all characters at a position

        :param int pos: position to return (0-based)

        :return str: all characters occuring at the position
        """
        return ''.join(seq[pos] for seq in self._seqs)

    def get_seq_len(self):
        "Get the length of the alignment"
        return len(self._seqs[0])

    def get_consensus(self, nucl=True):
        """
        .. versionchanged:: 0.1.16
            added *nucl* parameter

        The consensus sequence is constructed by checking the nucleotide that
        has the maximum number of counts for each position in the alignment.

        Arguments:
            nucl (bool): specify if the alignment is nucleotidic

        Returns:
            str: consensus sequence
        """

        cons_seq = []
        nucs = REV_COMP.keys()

        for pos in range(0, self.get_seq_len()):
            site = self.get_position(pos)
            if nucl:
                nuc_pos = max(
                    ((site.count(nuc), nuc) for nuc in nucs),
                    key=lambda x: x[0]
                )[1]
            else:
                nuc_pos = max(collections.Counter(x for x in site if x != '-'))

            cons_seq.append(nuc_pos)
        return ''.join(cons_seq)

    def __iter__(self):

        for name, seq in zip(self._names, self._seqs):
            yield name, seq

    def get_snps(self, ref_seq=None, full_size=False):
        """
        A SNP is called for the nucleotide that has the most counts among the
        ones that differ in the each site of the alignment. If two nucleotides
        have the same maximum count, one is randomly chosen.

        :param str ref_seq: a reference sequence can be provided, if None, a
            consensus sequence is produced for the alignment
        :param bool full_size: if True a tuple is returned for each position in
            the alignment. If there is no SNP at a position the value for the
            SNP is None

        :return list: a list of tuples (position, SNP)
        """
        if ref_seq is None:
            ref_seq = self.get_consensus()

        nucs = REV_COMP.keys()
        snps = []

        for pos in range(0, self.get_seq_len()):
            ref_nuc = ref_seq[pos]

            alts = self.get_position(pos).replace(ref_nuc, '').replace('-', '')
            if not alts:
                if full_size:
                    snps.append((pos, None))
                continue
            alts = sorted(
                ((alts.count(nuc), nuc) for nuc in nucs),
                key=lambda x: x[0]
            )
            counts = [alt[0] for alt in alts]
            count_idx = counts.index(counts[-1])
            alt = random.choice(alts[count_idx:])
            snps.append((pos, alt[1]))

        return snps


def get_seq_number_of_syn(ref_seq, snps, start=0, trans_table=None):
    """
    Given a reference sequence and a list of SNPs, calculates the number of
    synonymous and non-synonymous SNP.

    :param str ref_seq: reference sequence
    :param iterable snps: list of tuples (position, SNP) - zero based index
    :param int start: the frame used for the reference {0, 1, 2}
    :param dict trans_table: translation table used - codon->AA

    :return tuple: synonymous and non-synonymous counts
    """
    if trans_table is None:
        trans_table = TRANS_TABLE

    syn = 0
    nonsyn = 0

    for idx in range(start, len(ref_seq), 3):
        ref_codon = ref_seq[idx:idx+3]
        end = idx + 2
        for pos, change in snps:
            if not between(pos, idx, end):
                continue
            change_pos = pos-idx
            codon = ref_codon[:change_pos] + change + ref_codon[change_pos+1:]

            if len(codon) < 3:
                continue

            if trans_table[codon] == trans_table[ref_codon]:
                syn += 1
            else:
                nonsyn += 1

    return syn, nonsyn


def check_snp_in_seq(ref_seq, pos, change, start=0, trans_table=None):
    """
    Check a SNP in a reference sequence if it is a synonymous or non-synonymous
    change.

    :param str ref_seq: reference sequence
    :param int pos: SNP position - it is expected to be a 1 based index
    :param str change: nucleotide change occuring at *pos*
    :param int start: the starting position for the coding region - 0 based
        index
    :param dict trans_table: translation table used - codon->AA

    :return bool: True if it is a synonymous change, False if non-synonymous
    """
    # pos is expected to be a 1 based index
    # returns true if snp is synonymous
    # start is expected to be 0 based
    if trans_table is None:
        trans_table = TRANS_TABLE

    codon_idx = ((pos - start) // 3) - (1 if (pos - start) % 3 == 0 else 0)

    codon_pos = codon_idx * 3 + start

    ref_codon = ref_seq[codon_pos:codon_pos + 3]
    snp_codon = list(ref_codon)
    snp_codon[(pos % 3) - 1] = change
    snp_codon = ''.join(snp_codon)

    return trans_table[snp_codon] == trans_table[ref_codon]


def sequence_gc_ratio(sequence):
    """
    .. versionadded:: 0.1.13

    Calculate GC ratio information for a sequence. The formula is:

    .. math::
        :label: gc_ratio

        \\frac {(A + T)}{(G + C)}

    Arguments:
        sequence (str): sequence

    Returns:
        float: GC ratio, or `numpy.nan` if G = C = 0
    """

    at_sum = (sequence.count('A') + sequence.count('T'))
    gc_sum = (sequence.count('G') + sequence.count('C'))

    try:
        return at_sum / gc_sum
    except ZeroDivisionError:
        return numpy.nan


def sequence_gc_content(sequence):
    """
    .. versionchanged:: 0.3.3
        in case of `ZeroDivisionError` returns .5

    .. versionadded:: 0.1.13

    Calculate GC content information for an annotation. The formula is:

    .. math::
        :label: gc_content

        \\frac {(G + C)}{(G + C + A + T)}


    Arguments:
        sequence (str): sequence

    Returns:
        float: GC content
    """

    at_sum = (sequence.count('A') + sequence.count('T'))
    gc_sum = (sequence.count('G') + sequence.count('C'))

    try:
        return gc_sum / (gc_sum + at_sum)
    except ZeroDivisionError:
        return .5


def sequence_composition(sequence, chars=tuple(REV_COMP)):
    """
    .. versionadded:: 0.1.13

    Returns the number of occurences of each unique character in the sequence

    Arguments:
        sequence (str): sequence
        chars (iterable, None): iterable of the chars to test, default to
            (A, C, T, G). if None checks all unique characters in the sequencce

    Yields:
        tuple: the first element is the nucleotide and the second is the number
        of occurences in *sequence*
    """
    if chars is None:
        chars = set(sequence)

    for nuc in chars:
        yield nuc, sequence.count(nuc)


def get_contigs_info(file_name, pp=False):
    """
    .. versionchanged:: 0.2.4
       file_name can be a *dict* name->seq or a list of sequences

    .. versionadded:: 0.2.1

    Given a file name for a fasta file with sequences, a dictionary of
    name->seq, or a list of sequences, returns the following information in a
    tuple, or a string if *pp* is True:

    - number of sequences
    - total base pairs
    - max length
    - min length
    - average length
    - N50 statistic

    Arguments:
        file_name (str): fasta file to open
        pp (bool): if True, a formatted string is returned

    Returns:
        str, tuple: the returned value depends on the value of *pp*, if True a
        formatted string is returned, otherwise the tuple with all values is.
    """

    if isinstance(file_name, dict):
        seqs = list(viewvalues(file_name))
        file_name = 'dictionary'
    elif isinstance(file_name, list):
        seqs = list(file_name)
        file_name = 'list'
    else:
        seqs = list(seq for name, seq in fasta.load_fasta(file_name))

    lengths = numpy.array([len(x) for x in seqs])

    info = (
        len(seqs),
        lengths.sum(),
        lengths.max(),
        lengths.min(),
        lengths.mean(),
        calc_n50(lengths),
    )

    if pp:
        info_str = '{}: {} contigs, {} bp, max {}, min {}, avg {:.2f}, n50 {}'
        info = info_str.format(
            file_name, *info
        )

    return info


def _sliding_window(seq, size, step=None):
    """
    .. versionadded:: 0.2.6

    Returns a generator, with every iteration yielding a subsequence of size
    *size*, with a step of *step*.

    Arguments:
        seq (str): sequnece
        size (int): size of the sliding window
        step (int, None): the step to use in the sliding window. If *None*,
            half of the sequence length is used

    Yields:
        str: a subsequence of size *size* and step *step*
    """
    for index in range(0, len(seq) - size + 1, size // 2 if step is None else step):
        yield seq[index:index+size]


def _get_kmers(seq, k):
    """
    .. versionadded:: 0.2.6

    Returns a generator, with every iteration yielding a kmer of size *k*

    Arguments:
        seq (str): sequence
        k (int): kmer size

    Yields:
        str: a portion of *seq*, of size *k* with a step of *1*
    """
    for index in range(0, len(seq) - k + 1):
        yield seq[index:index+k]


def _sequence_signature(seq, w_size, k_size=4, step=None):
    """
    .. versionadded:: 0.2.6

    Returns the signature of a sequence, based on a kmer length, over a sliding
    window. Each sliding window signature is placed in order into a list, with
    each element being a :class:`collections.Counter` instance whose keys are
    the kmer found in that window.

    Arguments:
        seq (str): sequence for which to get the signature
        w_size (int): size of the sliding window size
        k_size (int): size of the kmer to use :func:`get_kmers`
        step (int): step to use in :func:`sliding_window`

    Returns:
        list: a list of :class:`collections.Counter` instances, for each
        window used
    """
    kmer_counts = []
    for subseq in _sliding_window(seq, w_size, step):
        kmer_counts.append(
            collections.Counter(kmer for kmer in _get_kmers(subseq, k_size) if 'N' not in kmer)
        )
    return kmer_counts


def _signatures_matrix(seqs, w_size, k_size=4, step=None):
    """
    .. versionadded:: 0.2.6

    Return a matrix (pandas.DataFrame) where the columns are the kmer found in
    all sequences *seqs* and the rows are the a MultiIndex with the first level
    being the sequnce name and the second the index of the sliding window for
    which a signature was computed.

    Arguments:
        seqs (iterable): iterable that yields a tuple, with the first element
            being the sequence name and the second the sequence itself
        w_size (int): size of the sliding window size
        k_size (int): size of the kmer to use :func:`get_kmers`
        step (int): step to use in :func:`sliding_window`, defaults to half of
            the window size

    Returns:
        pandas.DataFrame: a DataFrame where the columns are the kmers and the
        rows are the signatures of each contigs/windows.
    """

    def flatten_contigs(data):
        for name, windows in viewitems(data):
            for index, window in enumerate(windows):
                yield (name, index), dict(window)

    step = w_size // 2 if step is None else step

    kmer_counts = {}

    for name, seq in seqs:
        sign = sequence_signature(seq, w_size, k_size, step=step)
        kmer_counts[name] = sign

    return pandas.DataFrame(
        dict(
            flatten_contigs(kmer_counts)
        )
    ).T.fillna(0)


def random_sequences_codon(n=1, length=150, codons=list(UNIVERSAL.keys()),
                           p=None, frame=None):
    """
    .. versionadded:: 0.3.3

    Returns an iterator of nucleotidic sequences, based on a defined genetic
    code (passed as parameter, defaults to the universal one). The sequence if
    first sampled with replacement from the codon list, with a number of codons
    that covers the length chosen plus an additional one to allow a frame shift
    as set by *frame*

    .. note::

        If the probability (for each codon) are supplied, the number of
        sequences required to match those probabilities within a 10% margin of
        error is of at least 10.000 sequences, for 5% at leas 100.000

    Arguments:
        n (int): number of sequences to yield
        length (int): length of the sequences
        codons (iterable): codons used when generating the sequences
        p (tuple): probability of each codon occurence, in the same order as
            *codons*
        frame (int or None): used to define a specific frame shift occuring in
            the sequence (0 to 2) or a random one (if *None*)

    Yields:
        str: string representing a nucleotidic sequence
    """
    sample_size = (length // 3) + (1 if length % 3 != 0 else 0) + 1
    codons = numpy.array(codons)

    if frame is not None:
        if (frame < 0) or (frame > 2):
            frame = 0
        sframe = frame

    for i in range(n):

        if frame is None:
            sframe = numpy.random.randint(0, 4)

        yield ''.join(
            numpy.random.choice(codons, size=sample_size, replace=True, p=p)
        )[sframe:length+sframe]


def random_sequences(n=1, length=150, p=None):
    """
    .. versionadded:: 0.3.3

    Returns an iterator of random squences, where each nucleotide probability
    can be customised in the order (A, C, T, G)

    Arguments:
        n (int): number of sequences to yield
        length (int): length of each sequence
        p (tuple): tuple with the probability of a nucleotide to occur, in the
            order A, C, T, G

    Yields:
        str: string representing a nucleotidic sequence
    """
    nucl = numpy.array(['A', 'C', 'T', 'G'])

    for i in range(n):

        yield ''.join(
            numpy.random.choice(nucl, size=length, replace=True, p=p)
        )


def qualities_model_decrease(length=150, scale=None, loc=35):
    """
    .. versionadded:: 0.3.3

    The model is a decreasing one, from 35 and depends on the length of the
    sequence.

    Arguments:
        length (int): length of the qualities
        scale (float): base level of the qualities
        loc (float): loc parameter of the normal distribution

    Return:
        tuple: first element is sequence qualities, the second element contains
        the distribution used to randomise them
    """
    if scale is None:
        scale = 2. / (length / 150)
    return loc - (numpy.log(numpy.arange(length) + 1) * (length / 150)), stats.norm.freeze(loc=0, scale=scale)


def qualities_model_constant(length=150, scale=1, loc=35):
    """
    .. versionadded:: 0.3.3

    Model with constant quality

    Arguments:
        length (int): length of the qualities
        scale (float): base level of the qualities
        loc (float): loc parameter of the normal distribution

    Return:
        tuple: first element is sequence qualities, the second element contains
        the distribution used to randomise them
    """
    return numpy.ones(length) * loc, stats.norm.freeze(scale=scale)


def extrapolate_model(quals, frac=.5, scale_adj=.5):
    """
    .. versionadded:: 0.3.3

    .. versionchanged:: 0.4.4
        returns now the minimum and maximum quality scores found, along with
        some messages for the stages of the function

    Extrapolate a quality model from a list of qualities. It uses internally
    a LOWESS as the base, which is used to estimate the noise as a normal
    distribution.

    Arguments:
        quals (list): list of arrays of qualities, sorted by position in the
            corresponding sequence
        frac (float): fraction of the data used for the LOWESS fit (uses
            statsmodels)
        scale_adj (float): value by which the scale of the normal distribution
            will be multiplied. Defaults to halving the scale

    Returns:
        tuple: the first element is the qualities fit with a LOWESS, the second
        element is the distribution, the third is the minimum quality score and
        the fourth element is the maximum quality score found in the sequences
    """
    if not isinstance(quals, list):
        quals = list(quals)

    endog = numpy.hstack(quals)
    min_qual = endog.min()
    max_qual = endog.max()
    LOG.info("Max: %d Min: %d", max_qual, min_qual)
    exog = numpy.hstack(
        [numpy.arange(len(qual)) + 1 for qual in quals]
    )

    LOG.info("Starting LOWESS fit")

    lowess = sm.nonparametric.lowess(endog, exog, frac=frac)
    lowess = interpolate.interp1d(
        lowess[:, 0],
        lowess[:, 1],
        kind='nearest',
        bounds_error=False,
        fill_value='extrapolate'
    )
    lowess = lowess(numpy.arange(exog.max()) + 1)

    LOG.info("Starting normal distribution fit")

    dist = numpy.hstack([
        qual - lowess[:len(qual)]
        for qual in quals
    ])
    dist_args = stats.norm.fit(dist)
    dist_args = (dist_args[0], dist_args[1] * scale_adj)
    dist = stats.norm(*dist_args)
    return lowess, dist, min_qual, max_qual


def random_qualities(n=1, length=150, model=None, max_qual=60, min_qual=0):
    """
    .. versionadded:: 0.3.3

    .. versionchanged: 0.4.4
        added *max_qual* and *min_qual* parameters, also the max quality value
        output is 60 now

    Arguments:
        n (int): number of quality arrays to yield
        length (int): length of the quality array
        model (tuple): a tuple specifying the qualities and error distribution,
            if *None* :func:`qualities_model_decrease` is used
        max_qual (int): max quality values allowed, since rarely the quality
            exceed 60 in sequencing data

    Yields:
        numpy.array: numpy array of qualities, with the maximum value of 60
        by default, or *max_qual*, same with *min_qual*
    """
    if model is None:
        model = qualities_model_decrease(length=length)

    base, dist = model
    num_high_qual = 0
    num_low_qual = 0

    for x in range(n):
        qual = numpy.round(base + dist.rvs(size=length)).astype(int)
        if qual.max() > max_qual:
            qual[qual > max_qual] = max_qual
            num_high_qual += 1
        if qual.min() < min_qual:
            qual[qual < min_qual] = min_qual
            num_low_qual += 1
        yield qual

    LOG.debug('Number of quality sequences over %d was %d out of %d', max_qual,
              num_high_qual, n)
    LOG.debug('Number of quality sequences below %d was %d out of %d',
              min_qual, num_low_qual, n)



"""
Functions used in SNPs manipulation
"""
from builtins import zip
import logging
import itertools
import functools
import csv
import copy
from future.utils import viewitems, viewvalues
from .filter import pipe_filters

import pandas
import numpy
import scipy.stats

LOG = logging.getLogger(__name__)


def build_rank_matrix(dataframe, taxonomy=None, taxon_rank=None):
    """
    Make a rank matrix from a :class:`pandas.Series` with the pN/pS values of a
    dataset.

    :param dataframe: :class:`pandas.Series` instance with a MultiIndex
        (gene-taxon)
    :param taxonomy: :class:`taxon.Taxonomy` instance with the full
        taxonomy
    :param str taxon_rank: taxon rank to limit the specifity of the taxa
        included

    :return: :class:`pandas.DataFrame` instance
    """
    if taxon_rank is None:
        taxon_index = sorted(set(dataframe.index.get_level_values('taxon')))
    else:
        taxon_index = sorted(
            set(
                taxonomy.get_ranked_taxon(taxon_id, taxon_rank).taxon_id
                for taxon_id in set(dataframe.index.get_level_values('taxon'))
                if taxonomy.get_ranked_taxon(
                    taxon_id, taxon_rank
                ).rank == taxon_rank
            )
        )

    gene_index = sorted(set(dataframe.index.get_level_values('gene')))

    # print taxon_index
    rank_matrix = pandas.DataFrame(index=gene_index, columns=taxon_index)

    for gene_id in rank_matrix.index:
        gene_array = dataframe.loc[gene_id]
        # print gene_id, type(gene_array)
        iterator = zip(gene_array.index, scipy.stats.rankdata(gene_array))
        for taxon_id, rank in iterator:
            if taxon_rank is not None:
                taxon_id = taxonomy.get_ranked_taxon(
                    taxon_id, taxon_rank
                ).taxon_id
                if taxonomy[taxon_id].rank != taxon_rank:
                    continue

            old_value = rank_matrix.get_value(gene_id, taxon_id)
            if numpy.isnan(old_value) or old_value > rank:
                rank_matrix.set_value(gene_id, taxon_id, rank)

    return rank_matrix


def group_rank_matrix(dataframe, gene_map):
    """
    Group a rank matrix using a mapping, in the form map_id->ko_ids.

    :param dataframe: instance of a rank matrix from :func:`build_rank_matrix`
    :param dict gene_map: dictionary with the mapping

    :return: :class:`pandas.DataFrame` instance
    """
    rank_matrix = pandas.DataFrame(
        index=list(gene_map.keys()),
        columns=dataframe.columns
    )

    for mapping_id, gene_ids in viewitems(gene_map):
        mapped_matrix = dataframe.loc[gene_ids]
        # we only use the minimum rank among the genes with a set function
        # min() will return only those
        for taxon_id, rank in mapped_matrix.mean().dropna().items():
            rank_matrix.set_value(mapping_id, taxon_id, rank)

    return rank_matrix


def write_sign_genes_table(out_file, dataframe, sign_genes, taxonomy,
                           gene_names=None):
    """
    Write a table with the list of significant genes found in a dataframe, the
    significant gene list is the result of
    :func:`wilcoxon_pairwise_test_dataframe`.

    :out_file: the file name or file object to write the file
    :dataframe: the dataframe which was tested for significant genes
    :sign_genes: gene list that are significant
    :taxonomy: taxonomy object
    :gene_names: dictionary with the name of the the genes. Optional
    """
    if isinstance(out_file, str):
        out_file = open(out_file, 'w')

    out_file = csv.writer(out_file)

    taxon_ids = sorted(set(dataframe.index.get_level_values('taxon')),
                       key=lambda x: taxonomy[x].s_name)

    genes = set(dataframe.index.get_level_values('gene'))

    header = ['gene_id', 'gene_name', 'significant', 'number of taxa']

    for taxon_id in taxon_ids:
        header.append(taxonomy[taxon_id].s_name + '_mean')
        header.append(taxonomy[taxon_id].s_name + '_stdev')

    out_file.writerow(header)

    for gene in genes:
        values = [''] * (len(taxon_ids) * 2)

        dataframe_gene = dataframe.loc[gene]

        for taxon_id, series in dataframe_gene.iterrows():
            value_idx = taxon_ids.index(taxon_id) * 2

            values[value_idx] = series.mean()
            values[value_idx + 1] = series.std()

        out_file.writerow(
            [
                gene,
                '' if gene_names is None else gene_names[gene],
                'Y' if gene in sign_genes else 'N',
                len(dataframe_gene)
            ] + values
        )


def order_ratios(ratios, aggr_func=numpy.median, reverse=False,
                 key_filter=None):
    """
    Given a dictionary of id->iterable where iterable contains the values of
    interest, the function uses aggr_func to sort (ascending by default) it and
    return a list with the key in the sorted order.

    :param dict ratios: dictionary instance id->iterable
    :param function aggr_func: any function returning a value that can be used
        as a key in sorting
    :param bool reverse: the default is ascending sorting (False), set to True
        to reverse key_filter: list of keys to use for ordering, if None, every
        key is used

    :return: iterable with the sort order
    """

    if key_filter is None:
        key_filter = list(ratios.keys())

    order = [
        (
            aggr_func(
                [ratio for ratio in ratios[key] if not numpy.isnan(ratio)]
            ) if [ratio for ratio in ratios[key] if not numpy.isnan(ratio)]
            else 0,
            key
        )
        for key in key_filter
    ]

    order.sort(key=lambda x: x[0], reverse=reverse)
    # print order

    # for x, y in order: print y, x, ratios[y]

    return [x[1] for x in order]


def combine_sample_snps(snps_data, min_num, filters, index_type=None,
                        gene_func=None, taxon_func=None, use_uid=False,
                        flag_values=False, haplotypes=True, store_uids=False,
                        partial_calc=False, partial_syn=True):
    """
    .. versionchanged:: 0.2.2
        added *use_uid* argument

    .. versionchanged:: 0.3.1
        added *haplotypes*

    .. versionchanged:: 0.4.0
        added *store_uids*
    
    .. versionchanged:: 0.5.3
        added *partial_calc* and *partial_type*

    Combine a dictionary sample->gene_index->GeneSyn into a
    :class:`pandas.DataFrame`. The dictionary is first filtered with the
    functions in `filters`, mapped to different taxa and genes using
    `taxon_func` and `gene_func` respectively. The returned DataFrame is also
    filtered for each row having at least a `min_num` of not NaN values.

    Arguments:
        snps_data (dict): dictionary with the `GeneSNP` instances
        min_num (int): the minimum number of not NaN values necessary in a row
            to be returned
        filters (iterable): iterable containing filter functions, a list can be
            found in :mod:`mgkit.snps.filter`
        index_type (str, None): if `None`, each row index for the DataFrame
            will be a MultiIndex with `gene` and `taxon` as elements. If the
            equals 'gene', the row index will be gene based and if 'taxon' will
            be taxon based
        gene_func (func): a function to map a gene_id to a gene_map. See
            :func:`.mapper.map_gene_id` for an example
        taxon_func (func): a function to map a taxon_id to a list of IDs. See
            :mod:`.mapper.map_taxon_id_to_rank` or
            :mod:`.mapper.map_taxon_id_to_ancestor` for examples
        use_uid (bool): if True, uses the `GeneSNP.uid` instead of
            `GeneSNP.gene_id`
        flag_values (bool): if True,
            :meth:`mgkit.snps.classes.GeneSNP.calc_ratio_flag` will be used,
            instead of :meth:`mgkit.snps.classes.GeneSNP.calc_ratio`
        haplotypes (bool): if *flag_values* is False, and *haplotypes* is
            True, the 0/0 case will be returned as 0 instead of NaN
        store_uids (bool): if True a dictionary with the uid used for each
            cell (e.g. gene/taxon/sample)
        partial_calc (bool): if True, only pS or pN values will be calculated,
            depending on the value of *partial_syn*
        partial_syn (bool): if both *partial_calc* and this are True, only pS
            values will be calculated. If this parameter is False, pN values
            will be calculated

    Returns:
        DataFrame: :class:`pandas.DataFrame` with the pN/pS values for the
        input SNPs, with the columns being the samples. if *store_uids* is True
        the return value is a tuple (DataFrame, dict)

    """
    sample_dict = dict((sample, {}) for sample in snps_data)
    multi_index = set()

    if gene_func is None:
        gene_func = functools.partial(itertools.repeat, times=1)
    if taxon_func is None:
        taxon_func = functools.partial(itertools.repeat, times=1)

    if store_uids:
        uids_dict = {}
    
    if flag_values:
        calc_method = 'calc_ratio_flag'
    if partial_calc:
        if partial_syn:
            calc_method = 'calc_ps'
        else:
            calc_method = 'calc_pn'
        

    for sample, genes_dict in viewitems(snps_data):

        LOG.info('Analysing SNP from sample %s', sample)

        for gene_syn in pipe_filters(viewvalues(genes_dict), *filters):

            iter_func = itertools.product(
                gene_func(gene_syn.uid if use_uid else gene_syn.gene_id),
                taxon_func(gene_syn.taxon_id)
            )

            for gene_id, taxon_id in iter_func:

                if index_type == 'gene':
                    key = gene_id
                elif index_type == 'taxon':
                    key = taxon_id
                else:
                    key = (gene_id, taxon_id)

                if store_uids:
                    try:
                        uids_dict[key].add(gene_syn.uid)
                    except KeyError:
                        uids_dict[key] = set([gene_syn.uid])

                # we don't care about info about ids and so on, only syn/nonsyn
                # and coverage, to use the calc_ratio method
                try:
                    sample_dict[sample][key].add(gene_syn)
                except KeyError:
                    # Needed with the new GeneSNP, because copies the
                    # references and the number of SNPs raises (the original
                    # data structure is modified)
                    sample_dict[sample][key] = copy.deepcopy(gene_syn)

                multi_index.add(key)

    if index_type is None:
        multi_index = pandas.MultiIndex.from_tuples(
            sorted(multi_index),
            names=('gene', 'taxon')
        )
    else:
        multi_index = pandas.Index(multi_index)

    # we already satisfied a minimum coverage filter or at least if doesn't
    # matter in the calculation anymore, using haplotypes=True, the special
    # case where syn=nonsyn=0 will result in a 0 as pN/pS for a GeneSyn
    # instance
    sample_dict = dict(
        (
            sample,
            dict(
                (key, getattr(gene, calc_method)() if (flag_values or partial_calc) else gene.calc_ratio(haplotypes=haplotypes))
                for key, gene in viewitems(row_dict)
            )
        )
        for sample, row_dict in viewitems(sample_dict)
    )
    dataframe = pandas.DataFrame(sample_dict, index=multi_index,
                                 columns=sorted(sample_dict.keys()))

    dataframe = dataframe[dataframe.count(axis=1) >= min_num]
    if store_uids:
        return (dataframe, uids_dict)
    else:
        return dataframe


def significance_test(dataframe, taxon_id1, taxon_id2,
                      test_func=scipy.stats.ks_2samp):
    """
    .. versionadded:: 0.1.11

    Perform a statistical test on each gene distribution in two different taxa.

    For each gene common to the two taxa, the distribution of values in all
    samples (columns) between the two specified taxa is tested.

    Arguments:
        dataframe: :class:`pandas.DataFrame` instance
        taxon_id1: the first taxon ID
        taxon_id2: the second taxon ID
        test_func: function used to test,
          defaults to :func:`scipy.stats.ks_2samp`

    Returns:
        pandas.Series: with all pvalues from the tests
    """
    LOG.info("Performing %s test", test_func.__name__)

    gene_ids = set(
        dataframe.select(
            lambda x: x[1] == taxon_id1
        ).index.get_level_values('gene') &
        dataframe.select(
            lambda x: x[1] == taxon_id2
        ).index.get_level_values('gene')
    )
    LOG.info("Number of genes to be tested: %d", len(gene_ids))

    sign_genes = {}

    for gene_id in gene_ids:
        tx1_vals = dataframe.loc[gene_id].loc[taxon_id1].dropna()
        tx2_vals = dataframe.loc[gene_id].loc[taxon_id2].dropna()

        pvalue = test_func(tx1_vals, tx2_vals)[1]

        sign_genes[gene_id] = pvalue

    return pandas.Series(sign_genes)


def flat_sample_snps(snps_data, min_cov):
    """
    .. versionadded:: 0.1.11

    Adds all the values of a gene across all samples into one instance of
    :class:`classes.GeneSNP`, giving the average gene among all samples.

    Arguments:
        snps_data (dict): dictionary with the instances of
            :class:`classes.GeneSNP`
        min_cov (int): minimum coverage required for the each instance to be
            added

    Returns:
        dict: the dictionary with only one key (`all_samples`), which can be
        used with :func:`combine_sample_snps`
    """
    samples = list(snps_data.keys())
    gene_ids = list(snps_data[samples[0]].keys())
    new_data = {'all_samples': {}}

    for gene_id in gene_ids:
        for sample in samples:
            gene_syn = snps_data[sample][gene_id]

            if gene_syn.coverage < min_cov:
                continue

            try:
                new_data['all_samples'][gene_id].add(gene_syn)
            except KeyError:
                new_data['all_samples'][gene_id] = copy.deepcopy(gene_syn)

    return new_data



import pytest
import mgkit.io.utils
from mgkit.io.utils import open_file

def test_open_file_text(tmpdir):
    test_string = 'test\n'
    file_name = tmpdir.join('test-open').strpath
    handle = open_file(file_name, mode='wb')
    handle.write(test_string.encode('ascii'))
    handle.close()
    assert open_file(file_name, mode='rb').read().decode('ascii') == test_string


def test_open_file_binary(tmpdir):
    test_string = b'test\n'
    file_name = tmpdir.join('test-open').strpath
    handle = open_file(file_name, mode='wb')
    handle.write(test_string)
    handle.close()
    assert open_file(file_name, mode='rb').read() == test_string


def test_open_file_gz(tmpdir):
    test_string = b'test\n'
    file_name = tmpdir.join('test-open.gz').strpath
    handle = open_file(file_name, mode='w')
    handle.write(test_string)
    handle.close()
    assert open_file(file_name, mode='r').read() == test_string


def test_open_file_bz2(tmpdir):
    test_string = b'test\n'
    file_name = tmpdir.join('test-open.bz2').strpath
    handle = open_file(file_name, mode='w')
    handle.write(test_string)
    handle.close()
    assert open_file(file_name, mode='r').read() == test_string


@pytest.mark.skipif(
    mgkit.io.utils.lzma is None,
    reason="lzma module not installed"
)
def test_open_file_xz(tmpdir):
    test_string = b'test\n'
    file_name = tmpdir.join('test-open.xz').strpath
    handle = open_file(file_name, mode='w')
    handle.write(test_string)
    handle.close()
    assert open_file(file_name, mode='r').read() == test_string



import pytest
from mgkit.db.mongo import GFFDB
from mgkit.io.gff import Annotation
from pymongo.errors import ServerSelectionTimeoutError
from pymongo import MongoClient
from uuid import uuid4

try:
    MongoClient(serverSelectionTimeoutMS=2).list_database_names()
    db_offline = False
except ServerSelectionTimeoutError:
    db_offline = True

@pytest.fixture
def connection():
    db_name = uuid4().hex
    collection = uuid4().hex
    data = GFFDB(db_name, collection, timeout=2)
    yield data
    data.conn.drop_database(db_name)

@pytest.fixture
def annotations():
    return [
        Annotation(seq_id='1', start=0, end=100, strand='+', uid='u1'),
        Annotation(seq_id='1', start=10, end=100, strand='-', uid='u2'),
        Annotation(seq_id='2', start=0, end=100, strand='+', uid='u3'),
    ]

skip_conn = pytest.mark.skipif(
    db_offline,
    reason='A running mongodb instance is required'
)

@skip_conn
def test_insert_one(annotations, connection):
    connection.insert_one(annotations[0])


@skip_conn
def test_insert_one(connection):
    with pytest.raises(TypeError):
        connection.insert_one('test')


@skip_conn
def test_insert_many(annotations, connection):
    connection.insert_many(annotations)


@skip_conn
def test_get_annotation1(annotations, connection):
    connection.insert_one(annotations[0])
    assert connection[annotations[0].uid] == annotations[0]


@skip_conn
def test_get_annotation2(annotations, connection):
    connection.insert_many(annotations)
    values = [
        connection[annotation.uid] == annotation
        for annotation in annotations
    ]


@skip_conn
def test_values(annotations, connection):
    connection.insert_many(annotations)
    assert {annotation.uid for annotation in annotations} == \
        {annotation.uid for annotation in connection.values()}


@skip_conn
def test_iter(annotations, connection):
    connection.insert_many(annotations)
    assert {annotation.uid for annotation in annotations} == \
        {annotation.uid for annotation in connection}


@skip_conn
def test_itervalues(annotations, connection):
    connection.insert_many(annotations)
    assert {annotation.uid for annotation in annotations} == \
        {annotation.uid for annotation in connection.itervalues()}


@skip_conn
def test_items(annotations, connection):
    connection.insert_many(annotations)
    assert {annotation for annotation in annotations} == \
        {annotation for uid, annotation in connection.items()}


@skip_conn
def test_iteritems(annotations, connection):
    connection.insert_many(annotations)
    assert {annotation for annotation in annotations} == \
        {annotation for uid, annotation in connection.iteritems()}


@skip_conn
def test_keys(annotations, connection):
    connection.insert_many(annotations)
    assert {annotation.uid for annotation in annotations} == \
        {uid for uid in connection.keys()}



"""
.. versionadded:: 0.1.13

Uniprot file formats
"""
import sys
import logging
from . import open_file

LOG = logging.getLogger(__name__)

NUM_LINES = 10 ** 7


MAPPINGS = {
    'taxonomy': 'NCBI_TaxID',
    'eggnog': 'eggNOG',
    'ko': 'KO',
    'kegg': 'KEGG',
    'biocyc': 'BioCyc',
    'unipathway': 'UniPathway',
    'embl': 'EMBL',
    'embl_cds': 'EMBL-CDS',
    'gi': 'GI',
    'string': 'STRING'
}
"""
Some of the mappings contained in the idmapping.dat.gz
"""


def parse_uniprot_mappings(file_handle, gene_ids=None, mappings=None,
                           num_lines=NUM_LINES):
    """
    Parses a Uniprot mapping `file <ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/idmapping.dat.gz>`_,
    returning a generator with the mappings.

    Arguments:
        file_handle (str, file): file name or open file handle
        gene_ids (None, set): if not None, the returned mappings are for the
            gene IDs specified
        mappings (None, set): mappings to be returned
        num_lines (None, int): number of which a message is logged. If None,
            no message is logged

    Yields:
        tuple: the first element is the gene ID, the second is the mapping type
        and third element is the mapped ID
    """

    file_handle = open_file(file_handle, 'rb')

    LOG.info(
        "Loading Uniprot Mappings from file (%s)",
        getattr(file_handle, 'name', repr(file_handle))
    )

    if gene_ids is not None:
        gene_ids = set(gene_ids)
        LOG.info("Mappings for %d gene_ids will be returned", len(gene_ids))

    if mappings is not None:
        mappings = set(mappings)
        LOG.info(
            "Mappings to '%s' will be returned",
            ', '.join(mappings)
        )

    for idx, line in enumerate(file_handle):
        line = line.decode('ascii')

        if (num_lines is not None) and ((idx + 1) % num_lines == 0):
            LOG.info("Parsed %d lines", idx + 1)

        gene_id, mapping, map_id = line.strip().split('\t')

        if (gene_ids is not None) and (gene_id not in gene_ids):
            continue

        if (mappings is not None) and (mapping not in mappings):
            continue

        yield gene_id, mapping, map_id

    LOG.info("Read %d lines", idx + 1)


def uniprot_mappings_to_dict(file_handle, gene_ids, mappings, num_lines=None):
    """
    .. versionchanged:: 0.3.4
        added *num_lines*

    Parses a Uniprot mapping `file <ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/idmapping.dat.gz>`_,
    returning a generator of dictionaries with the mappings requested.

    Arguments:
        file_handle (str, file): file name or open file handle
        gene_ids (None, set): if not None, the returned mappings are for the
            gene IDs specified
        mappings (None, set): mappings to be returned
        num_lines (int, None): passed to :func:`parse_uniprot_mappings`

    Yields:
        tuple: the first element is the gene ID, the second is a dictionary
        with all the mappings found, the key is the mapping type and the value
        is a list of all mapped IDs
    """
    iterator = parse_uniprot_mappings(
        file_handle,
        gene_ids=gene_ids,
        mappings=mappings,
        num_lines=num_lines
    )

    curr_gene = ''
    curr_maps = {}

    for gene_id, mapping, map_id in iterator:
        if curr_gene == gene_id:
            try:
                curr_maps[mapping].append(map_id)
            except KeyError:
                curr_maps[mapping] = [map_id]
        else:
            if curr_gene != '':
                yield curr_gene, curr_maps
                curr_maps = {}
            curr_gene = gene_id
            curr_maps[mapping] = [map_id]
    else:
        yield curr_gene, curr_maps



import pytest
import pathlib
from mgkit.io import gff, fasta
from mgkit.io import open_file
from mgkit.utils import sequence

@pytest.fixture
def nucseq(shared_datadir):
    return dict(
        fasta.load_fasta(str(shared_datadir / 'test-seq-nuc.fa'))
    )


@pytest.fixture
def aaseq(shared_datadir):
    return dict(
        fasta.load_fasta(str(shared_datadir / 'test-seq-aa.fa'))
    )


@pytest.fixture
def gff_file(shared_datadir):
    return open_file(str(shared_datadir / 'test.gff'), 'rb').readlines()


@pytest.fixture
def hmmer_file(shared_datadir):
    return open_file(str(shared_datadir / 'test-hmmer-dom.txt'), 'rb').readlines()


@pytest.fixture
def glimmer_file(shared_datadir):
    return open_file(str(shared_datadir / 'glimmer3.txt'), 'rb').readlines()


@pytest.fixture
def keggmod_file(shared_datadir):
    return open_file(str(shared_datadir / 'kmod-entry1.txt'), 'rb').readlines()


def test_fromgff1(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert "KMSRIGKLPITVPAGVTVTVDENNLVTVKGPKGTLSQQVNPDITLKQEGNILTLERPTDSKPHKAMHGL" ==  ann.attr['aa_seq']


def test_fromgff2(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert 3 == ann.start


def test_fromgff3(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert 209 == ann.end


def test_fromgff_nouid1(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert len(ann.uid) != len('K02933.12503')


def test_fromgff_nouid2(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert ann.uid != '32ea1cc8-9e76-4310-8d1c-8e7890734a6b'


def test_fromgff_uid1(gff_file):

    ann = gff.from_gff(gff_file[1])

    assert ann.uid == '32ea1cc8-9e76-4310-8d1c-8e7890734a6b'


def test_Annotation_dbq(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert ann.dbq == 10


def test_Annotation_ec1(gff_file):

    ann = gff.from_gff(gff_file[1])

    assert ann.get_ec() == set(['1.1.-', '2.2.3.4'])


def test_Annotation_ec2(gff_file):

    ann = gff.from_gff(gff_file[1])

    assert ann.get_ec(level=2) == set(['1.1', '2.2'])


def test_Annotation_ec2_dup(gff_file):

    ann = gff.from_gff(gff_file[1])
    ann.attr['EC'] = ann.attr['EC'] + ',2.2.3.1'

    assert ann.get_ec(level=2) == set(['1.1', '2.2'])


def test_Annotation_ec3(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert ann.get_ec() == set()


def test_Annotation_get_mapping1(gff_file):

    ann = gff.from_gff(gff_file[2])

    assert ann.get_mapping('test') == ['12345']


def test_Annotation_get_mapping2(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert ann.get_mapping('test') == []


def test_Annotation_get_mappings1():

    ann = gff.Annotation()
    ann.set_mapping('ko', ['test'])
    mappings = ann.get_mappings()
    assert mappings['ko'] == ['test']


def test_Annotation_set_mapping1():
    ann = gff.Annotation()
    ann.set_mapping('ko', ['test'])
    assert ann.attr['map_KO'] == 'test'


def test_Annotation_set_mapping2():
    ann = gff.Annotation()
    ann.set_mapping('ko', ['test1', 'test2'])
    assert ann.attr['map_KO'] == 'test1,test2'


def test_Annotation_taxon_id_get1():
    ann = gff.Annotation()
    assert ann.taxon_id is None


def test_Annotation_taxon_id_set1():
    ann = gff.Annotation()
    ann.taxon_id = '10'
    assert ann.taxon_id == 10


def test_Annotation_taxon_id_set2():
    ann = gff.Annotation()
    ann.taxon_id = 10
    assert ann.taxon_id == 10


def test_Annotation_taxon_id_set3():
    ann = gff.Annotation()
    ann.taxon_id = 10.0
    assert ann.taxon_id == 10


def test_Annotation_add_exp_syn_count(gff_file, nucseq):

    ann = gff.from_gff(gff_file[0])
    ann.add_exp_syn_count(nucseq['contig-1327918'])

    assert (141, 480) == (ann.exp_syn, ann.exp_nonsyn)


def test_Annotation_add_gc_content(gff_file, nucseq):

    ann = gff.from_gff(gff_file[0])
    ann.add_gc_content(nucseq['contig-1327918'])

    assert 0.5314009661835749 == ann.get_attr('gc_cont', float)


def test_Annotation_add_gc_ratio(gff_file, nucseq):

    ann = gff.from_gff(gff_file[0])
    ann.add_gc_ratio(nucseq['contig-1327918'])

    assert 0.8818181818181818 == ann.get_attr('gc_ratio', float)


def test_Annotation_to_gff(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert gff.from_gff(ann.to_gff()).attr == ann.attr


def test_Annotation_to_file(gff_file, tmpdir):

    ann = gff.from_gff(gff_file[0])

    file_name = (tmpdir / 'test-write.gff').strpath
    file_handle = open_file(file_name, 'wb')
    ann.to_file(file_handle)
    file_handle.close()

    ann2 = next(gff.parse_gff(file_name))

    assert ann == ann2


def test_Annotation_to_gtf1(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert gff.from_gff(ann.to_gtf()).attr['transcript_id'] == ann.uid


def test_Annotation_to_gtf2(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert gff.from_gff(ann.to_gtf(gene_id_attr='ko')).attr['transcript_id'] == ann.gene_id


def test_Annotation_sample_coverage(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert int(ann.attr['t1_b3_cov']) == ann.sample_coverage['t1_b3']


def test_Annotation_get_number_of_samples1(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert ann.get_number_of_samples(min_cov=0) == 14


def test_Annotation_get_number_of_samples2(gff_file):

    ann = gff.from_gff(gff_file[0])

    assert ann.get_number_of_samples(min_cov=15) == 8


def test_Annotation_get_nuc_seq1():
    ann = gff.Annotation(start=1, end=40, strand='+')
    seq = 'ACTG' * 10

    assert ann.get_nuc_seq(seq) == seq


def test_Annotation_get_nuc_seq2():
    ann = gff.Annotation(start=2, end=40, strand='+')
    seq = 'ACTG' * 10

    assert ann.get_nuc_seq(seq) == seq[1:]


def test_Annotation_get_nuc_seq3():
    ann = gff.Annotation(start=2, end=39, strand='+')
    seq = 'ACTG' * 10

    assert ann.get_nuc_seq(seq) == seq[1:-1]


def test_Annotation_get_nuc_seq_reverse1():
    ann = gff.Annotation(start=1, end=40, strand='-')
    seq = 'ACTG' * 10

    assert ann.get_nuc_seq(seq) == seq


def test_Annotation_get_nuc_seq_reverse2():
    ann = gff.Annotation(start=2, end=39, strand='-')
    seq = 'ACTG' * 10

    assert ann.get_nuc_seq(seq, reverse=True) == sequence.reverse_complement(seq[1:-1])


def test_Annotation_get_nuc_seq_snp():
    ann = gff.Annotation(start=2, end=39, strand='+')
    seq = 'ACTG' * 10

    ann.get_nuc_seq(seq, snp=(2, 'C')) == sequence.get_variant_sequence(seq[1:-1], (2, 'C'))


def test_Annotation_get_aa_seq():
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert ann.get_aa_seq(seq) == sequence.translate_sequence(
            ann.get_nuc_seq(seq, reverse=True),
            start=0,
            reverse=False
        )


def test_Annotation_get_aa_seq__start1():
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert ann.get_aa_seq(seq, start=None) == sequence.translate_sequence(
            ann.get_nuc_seq(seq, reverse=True),
            start=ann.phase,
            reverse=False
        )


def test_Annotation_get_aa_seq__start2():
    ann = gff.Annotation(start=1, end=40, strand='+', phase=1)
    seq = 'ACTG' * 10

    assert ann.get_aa_seq(seq, start=None) == sequence.translate_sequence(
            ann.get_nuc_seq(seq, reverse=True),
            start=ann.phase,
            reverse=False
        )


def test_Annotation_get_aa_seq__snp():
    ann = gff.Annotation(start=1, end=40, strand='+', phase=1)
    seq = 'ACTG' * 10

    assert ann.get_aa_seq(seq, start=None, snp=(1, 'C')) == sequence.translate_sequence(
            ann.get_nuc_seq(seq, reverse=True, snp=(1, 'C')),
            start=ann.phase,
            reverse=False
        )


def test_Annotation_is_syn1_3():
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert ann.is_syn(seq, 3, 'C')


def test_Annotation_is_syn1_1():
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 1, 'C')


def test_Annotation_is_syn2_1():
    # second position on reference, first base in codon
    ann = gff.Annotation(start=2, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 2, 'A')


def test_Annotation_is_syn2_2():
    # second position on reference, second base in codon
    ann = gff.Annotation(start=2, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 3, 'A')


def test_Annotation_is_syn2_3():
    # second position on reference, third base in codon
    ann = gff.Annotation(start=2, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert ann.is_syn(seq, 4, 'A')


def test_Annotation_is_syn3_1():
    # first position on reference, second codon, first base in codon
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 4, 'A')


def test_Annotation_is_syn3_2():
    # first position on reference, second codon, second base in codon
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 5, 'T')


def test_Annotation_is_syn3_3():
    # first position on reference, second codon, third base in codon
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert ann.is_syn(seq, 6, 'T')


def test_Annotation_is_syn4_1():
    # first position on reference, third codon, first base in codon
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 7, 'A')


def test_Annotation_is_syn4_2():
    # first position on reference, third codon, second base in codon
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 8, 'T')


def test_Annotation_is_syn4_3():
    # first position on reference, third codon, third base in codon
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 9, 'T')


def test_Annotation_is_syn__start1_1():
    # first position on reference, second codon, first base in codon, phase=1
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 5, 'G', start=1)


def test_Annotation_is_syn__start1_2():
    # first position on reference, second codon, second base in codon, phase=1
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 6, 'A', start=1)


def test_Annotation_is_syn__start1_3():
    # first position on reference, second codon, third base in codon, phase=1
    ann = gff.Annotation(start=1, end=40, strand='+', phase=0)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 6, 'A', start=1)


def test_Annotation_is_syn__start2_1():
    # first position on reference, second codon, first base in codon, phase=1
    ann = gff.Annotation(start=1, end=40, strand='+', phase=1)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 6, 'A', start=None)


def test_Annotation_is_syn__start3_1():
    # first position on reference, third codon, third base in codon, phase=1
    ann = gff.Annotation(start=1, end=40, strand='+', phase=1)
    seq = 'ACTG' * 10

    assert ann.is_syn(seq, 10, 'T', start=None)


def test_Annotation_is_syn__start3_2():
    # first position on reference, third codon, second base in codon, phase=1
    ann = gff.Annotation(start=1, end=40, strand='+', phase=1)
    seq = 'ACTG' * 10

    assert not ann.is_syn(seq, 9, 'C', start=None)


def test_from_hmmer(hmmer_file, aaseq):
    checks = (
        'contig-1442648',
        'K00001_4479_poaceae',
        693,
        894,
        'K00001',
        4479,
        'poaceae'
    )
    ann = gff.from_hmmer(
        hmmer_file[0], aaseq
    )

    assert (
            ann.seq_id,
            ann.attr['name'],
            ann.attr['aa_from'],
            ann.attr['aa_to'],
            ann.gene_id,
            ann.taxon_id,
            ann.attr['taxon_name']
        ) == checks


def test_gff_glimmer3_line1():
    header = 'sequence0001'
    line = 'orf00001       66      611  +3     6.08'
    annotation = gff.from_glimmer3(header, line)

    assert (
            annotation.seq_id,
            annotation.start,
            annotation.end,
            annotation.score,
            annotation.strand,
            annotation.phase,
            annotation.attr['orf_id'],
            annotation.attr['frame']
        ) == (
            header,
            66,
            611,
            6.08,
            '+',
            2,
            'orf00001',
            '+3'
        )


def test_gff_glimmer3_line2():
    header = 'sequence0001'
    line = 'orf00001       66      11  -2     6.08'
    annotation = gff.from_glimmer3(header, line)

    assert (
            annotation.start,
            annotation.end,
            annotation.strand,
            annotation.phase,
            annotation.attr['frame']
        ) == (
            11,
            66,
            '-',
            1,
            '-2'
        )


def test_gff_from_sequence1(nucseq):
    annotation = gff.from_sequence(
        'contig-110637',
        nucseq['contig-110637']
    )
    assert (
            annotation.seq_id,
            annotation.start,
            annotation.end,
        ) == (
            'contig-110637',
            1,
            len(nucseq['contig-110637'])
        )


def test_genomicrange_eq1():
    a = gff.GenomicRange(start=1)
    b = gff.GenomicRange(start=2)
    assert a != b


def test_genomicrange_eq2():
    a = gff.GenomicRange(seq_id='a')
    b = gff.GenomicRange(seq_id='b')
    assert a != b


def test_genomicrange_get_range():
    a = gff.GenomicRange(start=1, end=30)
    assert a.get_range() == (1, 30)


def test_genomicrange_str():
    a = gff.GenomicRange(seq_id='seq', start=1, end=20, strand='-')
    assert str(a) ==  "{0}({1}):{2}-{3}".format('seq', '-', 1, 20)


def test_genomicrange_repr():
    a = gff.GenomicRange(seq_id='seq', start=1, end=20, strand='-')
    assert repr(a) ==  "{0}({1}):{2}-{3}".format('seq', '-', 1, 20)


@pytest.mark.parametrize(
    "start,end,coords,result",
    [
        (20, 30, 1, False),
        (20, 30, 20, True),
        (20, 30, 30, True),
        (20, 30, 25, True),
        (20, 30, (25, 30), True),
        (20, 30, (25, 31), False),
        (20, 30, (19, 30), False),
        (20, 30, (30, 25), True),
    ]
)
def test_genomicrange_contains(start, end, coords, result):
    gen_range1 = gff.GenomicRange(start=start, end=end)

    assert (coords in gen_range1) == result


@pytest.mark.parametrize(
    "coords1,coords2,result",
    [
        ((20, 30), (25, 30), True),
        ((20, 30), (30, 25), True),
        ((20, 30), (19, 30), False),
        ((20, 30), (25, 31), False),
    ]
)
def test_genomicrange_contains(coords1, coords2, result):
    gen_range1 = gff.GenomicRange(start=coords1[0], end=coords1[1])
    gen_range2 = gff.GenomicRange(start=coords2[0], end=coords2[1])

    assert (gen_range2 in gen_range1) == result


@pytest.mark.parametrize(
    "coords1,coords2,result",
    [
        ((20, 30), (25, 30), True),
        ((20, 30), (30, 25), True),
        ((20, 30), (19, 30), False),
        ((20, 30), (25, 31), False),
    ]
)
def test_genomicrange_contains_annotation1(coords1, coords2, result):
    gen_range1 = gff.GenomicRange(start=coords1[0], end=coords1[1])

    assert (gff.Annotation(start=coords2[0], end=coords2[1]) in gen_range1) == result


def test_split_gff_file1(tmpdir, shared_datadir):
    gff_file = str(shared_datadir / 'test.gff')

    gff.split_gff_file(gff_file, (tmpdir / 'test{}.gff').strpath, 2)

    files = list(pathlib.Path(tmpdir.strpath).glob('*.gff'))

    assert len(files) == 2


def test_split_gff_file2(tmpdir, shared_datadir):

    gff_file = str(shared_datadir / 'test.gff')

    gff.split_gff_file(gff_file, (tmpdir / 'test{}.gff').strpath, 2)

    files = list(
        str(path)
        for path in pathlib.Path(tmpdir.strpath).glob('*.gff')
    )

    count1 = sum(1 for x in gff.parse_gff(gff_file))
    count2 = sum(1 for x in gff.parse_gff_files(files))

    assert count1 == count2


@pytest.mark.parametrize(
    "start1,end1,start2,end2,result",
    [
        (20, 30, 25, 30, True),
        (20, 30, 30, 25, True),
        (20, 30, 19, 30, False),
        (20, 30, 25, 31, False),
    ]
)
def test_genomicrange_contains_annotation(start1, end1, start2, end2, result):
    gen_range1 = gff.GenomicRange(start=start1, end=end1)

    assert (gff.Annotation(start=start2, end=end2) in gen_range1) == result


@pytest.mark.parametrize(
    "start,end,pos,relpos",
    [
        (20, 30, 20, 1),
        (20, 30, 30, 11),
        (20, 30, 25, 6),
    ]
)
def test_genomicrange_get_relative_pos(start, end, pos, relpos):
    gen_range1 = gff.GenomicRange(start=start, end=end)

    gen_range1.get_relative_pos(pos) == relpos


@pytest.mark.parametrize(
    "start,end,pos",
    [
        (20, 30, 19),
        (20, 30, 31),
    ]
)
def test_genomicrange_get_relative_pos_fail(start, end, pos):
    gen_range1 = gff.GenomicRange(start=start, end=end)

    with pytest.raises(ValueError):
        gen_range1.get_relative_pos(pos)


@pytest.mark.parametrize(
    "start2,end2,ustart,uend",
    [
        (19, 30, 10, 30),
        (10, 20, 10, 20),
        (20, 30, 10, 30),
        (21, 30, 10, 30),
    ]
)
def test_genomicrange_union1(start2, end2, ustart, uend):
    gen_range1 = gff.GenomicRange()
    gen_range1.seq_id = 'seq1'
    gen_range1.strand = '+'
    gen_range1.start = 10
    gen_range1.end = 20
    gen_range2 = gff.GenomicRange()
    gen_range2.seq_id = 'seq1'
    gen_range2.strand = '+'
    gen_range2.start = start2
    gen_range2.end = end2

    gen_range_u = gen_range1.union(gen_range2)

    assert (gen_range_u.start, gen_range_u.end) == (ustart, uend)


def test_genomicrange_union2():
    gen_range1 = gff.GenomicRange()
    gen_range1.seq_id = 'seq1'
    gen_range1.strand = '+'
    gen_range1.start = 10
    gen_range1.end = 20
    gen_range2 = gff.GenomicRange()
    gen_range2.seq_id = 'seq1'
    gen_range2.strand = '+'
    gen_range2.start = 21
    gen_range2.end = 30

    gen_range_u = gen_range1 | gen_range2

    assert (gen_range_u.start, gen_range_u.end) == (10, 30)


@pytest.mark.parametrize(
    "start2,end2,seq2,strand2",
    [
        (19, 30, 'seq2', '+'),
        (19, 30, 'seq1', '-'),
        (21, 30, 'seq1', '-'),
    ]
)
def test_genomicrange_union_fail(start2, end2, seq2, strand2):
    gen_range1 = gff.GenomicRange()
    gen_range1.seq_id = 'seq1'
    gen_range1.strand = '+'
    gen_range1.start = 10
    gen_range1.end = 20
    gen_range2 = gff.GenomicRange()
    gen_range2.seq_id = seq2
    gen_range2.strand = strand2
    gen_range2.start = start2
    gen_range2.end = end2

    gen_range_u = gen_range1.union(gen_range2)

    assert gen_range_u is None


@pytest.mark.parametrize(
    "start2,end2,seq2,strand2,result",
    [
        (19, 30, 'seq1', '+', (19, 20)),
        (15, 30, 'seq1', '+', (15, 20)),
        (10, 20, 'seq1', '+', (10, 20)),
    ]
)
def test_genomicrange_intersect1(start2, end2, seq2, strand2, result):
    gen_range1 = gff.GenomicRange()
    gen_range1.seq_id = 'seq1'
    gen_range1.strand = '+'
    gen_range1.start = 10
    gen_range1.end = 20
    gen_range2 = gff.GenomicRange()
    gen_range2.seq_id = seq2
    gen_range2.strand = strand2
    gen_range2.start = start2
    gen_range2.end = end2

    gen_range_u = gen_range1.intersect(gen_range2)

    assert (gen_range_u.start, gen_range_u.end) == result



def test_genomicrange_intersect2():
    gen_range1 = gff.GenomicRange(seq_id='seq1', strand='+', start=10, end=20)
    gen_range2 = gff.GenomicRange(seq_id='seq1', strand='+', start=12, end=18)

    gen_range_u = gen_range2.intersect(gen_range1)

    assert len(gen_range_u) == len(gen_range2)


def test_genomicrange_intersect3():
    gen_range1 = gff.GenomicRange(seq_id='seq1', strand='+', start=10, end=20)
    gen_range2 = gff.GenomicRange(seq_id='seq1', strand='+', start=12, end=18)

    gen_range_u = gen_range1.intersect(gen_range2)

    assert len(gen_range_u) == len(gen_range2)


def test_genomicrange_intersect4():
    gen_range1 = gff.GenomicRange(seq_id='seq1', strand='+', start=10, end=20)
    gen_range2 = gff.GenomicRange(seq_id='seq1', strand='+', start=12, end=18)

    gen_range_u = gen_range1 & gen_range2

    assert len(gen_range_u) == len(gen_range2)


@pytest.mark.parametrize(
    "seq1,strand1,s1,e1,seq2,strand2,s2,e2",
    [
        ('seq2', '+', 10, 20, 'seq1', '+', 10, 20),
        ('seq1', '-', 10, 20, 'seq1', '+', 10, 20),
        ('seq1', '+', 10, 20, 'seq1', '+', 30, 40),
    ]
)
def test_genomicrange_intersect_fail(seq1, strand1, s1, e1, seq2, strand2, s2, e2):
    gen_range1 = gff.GenomicRange()
    gen_range1.seq_id = seq1
    gen_range1.strand = strand1
    gen_range1.start = s1
    gen_range1.end = e1
    gen_range2 = gff.GenomicRange()
    gen_range2.seq_id = seq2
    gen_range2.strand = strand2
    gen_range2.start = s2
    gen_range2.end = e2

    gen_range_u = gen_range2.intersect(gen_range1)

    assert gen_range_u is None


@pytest.fixture
def elongate_data():
    seq_id = 'test1'

    test_ann = [
        gff.Annotation(seq_id=seq_id, start=1, end=10),
        gff.Annotation(seq_id=seq_id, start=10, end=15),
        gff.Annotation(seq_id=seq_id, start=16, end=18),
    ]
    return test_ann


def test_elongate_annotations(elongate_data):
    assert gff.elongate_annotations(elongate_data) == {(1, 18)}


def test_annotation_length1():
    ann = gff.Annotation(start=1, end=10)
    assert len(ann) == 10


def test_annotation_length2():
    ann = gff.Annotation(start=1, end=10)
    assert ann.length == 10


def test_genomicrange_length1():
    ann = gff.GenomicRange(start=1, end=10)
    assert len(ann) == 10


def test_genomicrange_length2():
    ann = gff.GenomicRange(start=1, end=10)
    with pytest.raises(AttributeError):
        ann.length



"""
Module containing constants for the filter package
"""

import mgkit.taxon

MIN_COV = 4
"Minumum coverage required in some functions."

MIN_NUM = 10
"Used to set the minimum number of replicates for some functions"

BLACK_LIST = [
    'bos',
    'pecora',
    'lolium',
    'streptophyta',
    'oryza',
    'fabales',
    'poaceae',
    'metazoa',
    'chlorophyta'
]
"""
Default taxa black list, includes all taxa names that are to be excluded from
some analysis.
"""

BLACK_LIST_IDS = [
    903,    # bos
    35500,  # pecora
    4520,   # lolium
    35493,  # streptophyta
    4527,   # oryza
    72025,  # fabales
    4479,   # poaceae
    33208,  # metazoa
    3041,   # chlorophyta
]

DEFAULT_SNP_FILTER = {
    'min_cov': MIN_COV,
    'black_list': BLACK_LIST_IDS,
    'include_only': [
        mgkit.taxon.ARCHAEA,
        mgkit.taxon.BACTERIA,
        mgkit.taxon.FUNGI,
    ] + list(mgkit.taxon.PROTISTS.values()),
    'func': mgkit.taxon.is_ancestor
}
"""
Default filter options for filtering :class:`mgkit.snps.GeneSyn`
"""



"""
.. versionadded:: 0.2.1

This module contains functions and classes to use for a DB like representation
of annotations using the *pymongo* package, a driver to **MongoDB**.

In a MongoDB document, exported from an annotation, using the
:meth:`mgkit.io.gff.Annotation.to_mongodb` method, the keys that are defined
are::

    seq_id, source, feat_type, start, end, score, strand,
    phase, gene_id, taxon_id, bitscore, exp_nonsyn, exp_syn,
    length, dbq, coverage, map

These are defined because they have values that are not strings (defined as
properties in :class:`mgkit.io.gff.Annotation`. The rest of the attributes
defined are kept as well, but no ckeck for the data type is made.

.. note::

    lineage is added as a key, whose values are taxon_id, if a function has
    been passed to :meth:`mgkit.io.gff.Annotation.to_mongodb`

The exception is the **map** key in the document. It store both the EC mappings
(EC attribute in the GFF), as well as all mappings whose attribute starts with
*map_*. The former is usually accessed from
:meth:`mgkit.io.gff.Annotation.get_ec` while the latter from
:meth:`mgkit.io.gff.Annotation.get_mapping` or
:meth:`mgkit.io.gff.Annotation.get_mappings`.

These 3 methods return a list and this list is used in the MongoDB document.
The MongoDB document will contain a **map** key where the values are the type
of mappings, and the values the list of IDs the annoation maps to.

.. list-table:: Example for the map dictionary
   :header-rows: 1

   * - Type
     - GFF
     - Annotation
     - MongoDB Document
     - MongoDB Query
   * - EC
     - EC
     - get_ec
     - ec
     - map.ec
   * - KO
     - map_KO
     - get_mapping('ko')
     - ko
     - map.ko
   * - eggNOG
     - map_EGGNOG
     - get_mapping('eggnog')
     - eggnog
     - map.eggnog


"""
from builtins import object
import logging
from ..io import gff

from pymongo import MongoClient

LOG = logging.getLogger(__name__)


class GFFDB(object):
    """
    .. versionchanged:: 0.3.4
        added *timeout* parameter

    Wrapper to a MongoDB connection/db. It is used to automate the convertion
    of MongoDB records into :class:`mgkit.io.gff.Annotation` instances.
    """
    conn = None
    db = None

    def __init__(self, db, collection, uri=None, timeout=5):
        self.conn = MongoClient(uri, serverSelectionTimeoutMS=timeout)
        self.db = self.conn[db][collection]

    def cursor(self, query=None):
        "Returns a cursor for the query"
        return self.db.find(query)

    def convert_record(self, record):
        """
        .. versionchanged:: 0.3.1
            removes *lineage* from the attributes

        Converts the record (a dictionary instance) to an Annotation
        """
        return gff.from_mongodb(record, lineage=False)

    def find_annotation(self, query=None):
        """
        Iterate over a cursor created using *query* and yields each record
        after converting it to a :class:`mgkit.io.gff.Annotation` instance,
        using :meth:`mgkit.db.mongo.GFFDB.convert_record`.
        """
        for record in self.cursor(query):
            yield self.convert_record(record)

    def __getitem__(self, uid):
        """
        .. versionadded:: 0.3.1

        Retrieves an annotation from the DB by its *uid*
        """
        return self.convert_record(self.db.find_one(uid))

    def insert_one(self, annotation):
        """
        .. versionadded:: 0.3.4

        Inserts an annotation into the DB

        Raises:
            TypeError: if the passed object is not an annotation
        """
        if not isinstance(annotation, gff.Annotation):
            raise TypeError(
                'An mgkit.io.gff.Annotation instance is required: {}'.format(
                    type(annotation)
                )
            )

        self.db.insert_one(annotation.to_mongodb(raw=True))

    def insert_many(self, annotations):
        """
        .. versionadded:: 0.3.4

        Inserts annotations into the DB

        .. warning::

            The object must be a :class:`mgkit.io.gff.Annotation`

        """

        self.db.insert_many(
            (
                annotation.to_mongodb(raw=True)
                for annotation in annotations
            ),
            ordered=False
        )

    def __iter__(self):
        """
        .. versionadded:: 0.3.1

        Iterates over all annotations
        """
        return self.values()

    def values(self):
        """
        .. versionadded:: 0.3.1

        Iterates over all the annotations in the db/collection
        """
        return self.find_annotation()

    def itervalues(self):
        """
        .. versionadded:: 0.3.1

        Alias for :meth:`GFFDB.values`
        """
        return self.values()

    def items(self):
        """
        .. versionadded:: 0.3.1

        Iterates over all the annotations in the db/collection, yielding a
        tuple (*annotation.uid*, *annotation*)
        """
        for record in self.values():
            yield record.uid, record

    def iteritems(self):
        """
        .. versionadded:: 0.3.1

        Alias for :meth:`GFFDB.items`
        """
        return self.items()

    def keys(self):
        """
        .. versionadded:: 0.3.1

        Iterates over all the *uid* in the db/collection
        """
        for record in self.values():
            yield record.uid



"""
This module gives access to Uniprot taxonomy data. It also defines classes
to filter, order and group data by taxa
"""
from builtins import object
from functools import reduce
import logging
import itertools
import collections
import pickle
from future.utils import viewitems, viewvalues
from .io import open_file


LOG = logging.getLogger(__name__)

ALGAE = {
    'haptophyceae': 2830,
    'chlorophyta': 3041,
    'stramenopiles': 33634,
    'cryptophyta': 3027,
    'rhodophyta': 2763,
}

# superkingdoms
BACTERIA = 2
ARCHAEA = 2157
VIRUS = 10239

CELLULAR_ORGANISMS = 131567

# kingdoms
FUNGI = 4751
VIRIDIPLANTAE = 33090
METAZOA = 33208
EUKARYOTA = 2759

PROTISTS = {
    # 'apicomplexa', alveolata
    # 'ciliophora', alveolata
    'alveolata': 33630,
    'amoebozoa': 554915,
    'apusozoa': 554296,
    'breviatea': 1401294,
    'centroheliozoa': 193537,
    'choanoflagellida': 28009,
    'diplomonadida': 5738,
    'euglenozoa': 33682,
    'formicata': 207245,
    'heterolobosea': 5752,
    'jakobida': 556282,
    'malawimonadidae': 136087,
    'oxymonadida': 66288,
    'parabasalia': 5719,
    'rhizaria': 543769,
}

PLANTS = {
    'streptophyta': 35493,
}

TAXON_ROOTS = (
    'archaea',  # 2157
    'bacteria',  # 2
    'fungi',  # 4751
    'metazoa',
    'environmental samples',
    'viruses',
    'viroids'
    'eukaryota',
    'other sequences',
    'unidentified'
) + tuple(PROTISTS) + tuple(PLANTS) + tuple(ALGAE)
"Root taxa used in analysis and filtering"

TAXON_RANKS = (
    'superkingdom',
    'kingdom',
    'phylum',
    'class',
    'subclass',
    'order',
    'family',
    'genus',
    'species'
)
"Taxonomy ranks included in the pickled data"

TaxonTuple = collections.namedtuple(
    'UniprotTaxonTuple',
    ('taxon_id', 's_name', 'c_name', 'rank', 'lineage', 'parent_id')
)
"""
A representation of a Uniprot Taxon
"""

# To be deprecated to avoid confusion
UniprotTaxonTuple = TaxonTuple


def parse_uniprot_taxon(line, light=True):
    """
    .. versionchanged:: 0.1.13
        now accepts empty scientific names, for root taxa

    .. versionchanged:: 0.2.1
        added *light* parameter

    .. deprecated:: 0.4.0

    Parses a Uniprot taxonomy file (tab delimited) line and returns a
    UniprotTaxonTuple instance. If *light* is True, lineage is not stored to
    decrease the memory usage. This is now the default.
    """
    line = line.decode('ascii')
    line = line.rstrip().split('\t')
    taxon_id = int(line[0])
    try:
        s_name = line[2].lower()
    except IndexError:
        s_name = ''
    try:
        c_name = line[3].lower() if line[3] else ''
    except IndexError:
        c_name = ''
    try:
        rank = line[7].lower()
    except IndexError:
        rank = None
    try:
        if light:
            raise IndexError
        lineage = tuple(x.lower() for x in line[8].split('; '))
    except IndexError:
        lineage = (None,)
    try:
        parent_id = int(line[9])
    except (ValueError, IndexError):
        parent_id = None
    return UniprotTaxonTuple(
        taxon_id,
        s_name,
        c_name,
        rank,
        lineage,
        parent_id
    )


def parse_ncbi_taxonomy_merged_file(file_handle):
    """
    .. versionadded:: 0.2.3

    Parses the *merged.dmp* file where the merged taxon_id are stored. Available
    at ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/

    Arguments:
        file_handle (str, file): file name or handle to the file

    Returns:
        dict: dictionary with merged_id -> taxon_id
    """
    file_handle = open_file(file_handle, 'r')

    LOG.info(
        "Reading NCBI taxonomy merged file %s",
        getattr(file_handle, 'name', repr(file_handle))
    )

    merged_taxa = {}

    for line in file_handle:
        line = line.decode('ascii')
        merged_id, taxon_id = [col for col in line.strip().split('\t') if col != '|']
        merged_taxa[int(merged_id)] = int(taxon_id)

    return merged_taxa


def parse_ncbi_taxonomy_names_file(file_handle, name_classes=('scientific name', 'common name')):
    """
    .. versionadded:: 0.2.3

    Parses the *names.dmp* file where the names associated to a taxon_id are
    stored. Available at ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/

    Arguments:
        file_handle (str, file): file name or handle to the file
        name_classes (tuple): name classes to save, only the scientific and
            common name are stored

    Returns:
        dict: dictionary with merged_id -> taxon_id
    """

    file_handle = open_file(file_handle, 'r')

    LOG.info(
        "Reading NCBI taxonomy names file %s",
        getattr(file_handle, 'name', repr(file_handle))
    )

    taxa_names = {}

    for line in file_handle:
        line = line.decode('ascii')
        taxon_id, taxon_name, uniq_name, name_class = [
            col.strip()
            for col in line.strip().split('\t')
            if col != '|'
        ]
        taxon_id = int(taxon_id)

        if name_class not in name_classes:
            continue

        if taxon_id not in taxa_names:
            taxa_names[taxon_id] = {}

        taxa_names[taxon_id][name_class] = taxon_name

    return taxa_names


def parse_ncbi_taxonomy_nodes_file(file_handle, taxa_names=None):
    """
    .. versionadded:: 0.2.3

    Parses the *nodes.dmp* file where the nodes of the taxonomy are stored.
    Available at ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/.

    Arguments:
        file_handle (str, file): file name or handle to the file
        taxa_names (dict): dictionary with the taxa names (returned from
            :func:`parse_ncbi_taxonomy_names_file`)

    Yields:
        TaxonTuple: TaxonTuple instance
    """

    file_handle = open_file(file_handle, 'r')

    LOG.info(
        "Reading NCBI taxonomy nodes file %s",
        getattr(file_handle, 'name', repr(file_handle))
    )

    for line in file_handle:
        line = line.decode('ascii')
        line = [col for col in line.strip().split('\t') if col != '|']
        taxon_id = int(line[0])
        parent_id = int(line[1])
        if parent_id == 1:
            # NCBI uses 1 as the root, but the rest of the functions expect None
            parent_id = None
        rank = line[2].lower()

        s_name = ''
        c_name = ''

        if taxa_names is not None:
            try:
                names = taxa_names[taxon_id]
            except KeyError:
                names = {}
            s_name = names.get('scientific name', '')
            c_name = names.get('common name', '')

        yield TaxonTuple(
            taxon_id,
            s_name,
            c_name,
            rank,
            (None,),  # lineage is not found in the NCBI taxonomy dump
            parent_id
        )


class Taxonomy(object):
    """
    Class that contains the whole Uniprot taxonomy. Defines some methods to
    easy access of taxonomy. Follows the conventions of NCBI Taxonomy.

    Defines:

    * methods to load taxonomy from a pickle file or a generic file handle
    * can be iterated over and returns a generator its UniprotTaxon instances
    * can be used as a dictionary, in which the key is a taxon_id and the value
      is its UniprotTaxon instance
    """
    def __init__(self, fname=None):
        """
        :param str fname: file name of the pickled data.
        """
        self._taxa = {}
        self._name_map = {}
        if fname:
            self.load_data(fname)

    @staticmethod
    def parse_gtdb_lineage(lineage, sep=';'):
        """
        .. versionadded:: 0.3.3

        Parse a GTDB lineage, one that defines the rank as a single letter,
        followed by **__** for each taxon name. Taxa are separated by semicolon
        by default. Also the **domain** rank is renamed into **superkingdom**
        to allow mixing of taxonomies.

        Returns:
            dict: dictionary with the parsed lineage, which can be passed to
            :meth:`Taxonomy.add_lineage`
        """
        lineage_dict = {}
        ranks = dict(d='superkingdom', f='family', p='phylum', o='order', g='genus', s='species', c='class')
        for taxon_name in lineage.split(sep):
            rank, taxon_name = taxon_name.split('__')
            if not taxon_name:
                continue
            lineage_dict[ranks[rank]] = taxon_name
        return lineage_dict

    def read_from_gtdb_taxonomy(self, file_handle, use_gtdb_name=True, sep='\t'):
        """
        .. versionadded:: 0.3.0

        .. versionchanged:: 0.3.1
            replaced *domain* with *superkingdom* to support *get_lineage*

        Reads a GTDB taxonomy file (tab separated genome_id/taxonomy) and
        populate the taxonomy instance. The method also return a dictionary of
        genome_id -> taxon_id.

        Arguments:
            file_handle (file): file with the taxonomy
            use_gtdb_name (bool): if True, the names are kept as-is in the
                *s_name* attribute of :class:`TaxonTuple` and the
                "cleaned" version in *c_name* (e.g. f__Ammonifexaceae ->
                Ammonifexaceae). If False, the values are switched
            sep (str): separator between the columns of the file

        Returns:
            dict: dictionary of genome_id -> taxon_id, reflecting the created
            taxonomy

        .. note::

            the taxon_id are generated, so there's no guarantee they will be
            the same in a successive execution

        """
        if isinstance(file_handle, str):
            file_handle = open_file(file_handle, 'r')

        LOG.info(
            "Reading GTDB taxonomy from file (%s)",
            getattr(file_handle, 'name', repr(file_handle))
        )

        genome_ids = {}

        # ranks used in GTDB
        ranks = {
            'g': 'genus',
            'f': 'family',
            'p': 'phylum',
            'o': 'order',
            'd': 'superkingdom',
            's': 'species',
            'c': 'class'
        }

        taxon_ids = {}
        count = 1
        for line in file_handle:
            line = line.decode('ascii')
            # expecting the table to be the exported file from GTDB
            try:
                # print line.strip().split(sep)
                genome_id, line = line.strip().split(sep)
                # in case there's no gtdb taxonomy
            except ValueError:
                continue

            line = [
                taxon_name.strip()
                for taxon_name in line.split(';')
                if len(taxon_name) > 3
            ]

            if not line:
                continue

            # if no parent, use None as the NCBI taxonomy
            parent_id = None
            for taxon_name in line:

                # register a taxo_id for unknown taxon
                if taxon_name not in taxon_ids:
                    taxon_ids[taxon_name] = count
                    count += 1

                taxon_id = taxon_ids[taxon_name]
                try:
                    rank = ranks[taxon_name[0]]
                except KeyError:
                    raise KeyError

                # keep the full gtdb name in the common name
                # cut the scientific name to remove the rank information
                common_name = taxon_name
                taxon_name = taxon_name[3:]
                # but if th use_gtdb_name option is True, switch them
                if use_gtdb_name:
                    taxon_name, common_name = common_name, taxon_name

                self[taxon_id] = TaxonTuple(
                    taxon_id,
                    taxon_name,
                    common_name,
                    rank,
                    (None,),
                    parent_id
                )
                parent_id = taxon_id
            genome_ids[genome_id] = taxon_id

        return genome_ids

    def read_from_ncbi_dump(self, nodes_file, names_file=None, merged_file=None):
        """
            .. versionadded:: 0.2.3

            Uses the *nodes.dmp* and optionally *names.dmp*, *merged.dmp* files
            from ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/ to populate the
            taxonomy.

            Arguments:
                nodes_file (str, file): file name or handle to the file
                names_file (str, file, None): file name or handle to the file,
                    if None, names won't be added to the taxa
                merged_file (str, file, None): file name or handle to the file,
                    if None, pointers to merged taxa won't be added
            """

        LOG.info("Reading NCBI taxonomy from file")

        if names_file is not None:
            taxa_names = parse_ncbi_taxonomy_names_file(names_file)
        else:
            taxa_names = None

        for taxon in parse_ncbi_taxonomy_nodes_file(nodes_file, taxa_names):
            self[taxon.taxon_id] = taxon

        if merged_file is not None:
            merged_taxa = parse_ncbi_taxonomy_merged_file(merged_file)

            for merged_id, taxon_id in viewitems(merged_taxa):
                self[merged_id] = self[taxon_id]

    def read_taxonomy(self, f_handle, light=True):
        """
        .. versionchanged:: 0.2.1
            added *light* parameter

        .. deprecated:: 0.4.0
            use :meth:`Taxonomy.read_from_ncbi_dump`

        Reads taxonomy from a file handle.
        The file needs to be a tab separated format return by a query on
        Uniprot.  If *light* is True, lineage is not stored to decrease the
        memory usage. This is now the default.

        New taxa will be added, duplicated taxa will be skipped.

        :param handle f_handle: file handle of the taxonomy file.
        """
        try:
            LOG.info("Loading taxonomy from file %s", f_handle.name)
        except AttributeError:
            pass

        f_handle.readline()

        for line in f_handle:
            try:
                # taxon = UniprotTaxon(line=line)
                taxon = parse_uniprot_taxon(line, light=light)
            except Exception:
                # print line.strip().split('\t')
                LOG.debug("skipped line: %s", line)
                continue
            if taxon.taxon_id in self:
                continue
            self[taxon.taxon_id] = taxon

    def gen_name_map(self):
        """
        .. versionchanged:: 0.2.3
            names are stored in the mapping as lowercase

        Generate a name map, where to each scientific name in the taxonomy an
        id is associated.
        """
        LOG.debug('Generate name map')
        self._name_map = {}
        for taxon_obj in self:
            s_name = taxon_obj.s_name.lower()
            try:
                self._name_map[s_name].append(taxon_obj.taxon_id)
            except KeyError:
                self._name_map[s_name] = [taxon_obj.taxon_id]

    def load_data(self, file_handle):
        """
        .. versionchanged:: 0.2.3
            now can use read *msgpack* serialised files

        .. versionchanged:: 0.1.13
            now accepts file handles and compressed files (if file names)

        Loads serialised data from file name "file_handle" and accepts
        compressed files.

        if the *.msgpack* string is found in the file name, the *msgpack*
        package is used instead of pickle

        Arguments:
            file_handle (str, file): file name to which save the instance data

        """

        if isinstance(file_handle, str):
            file_handle = open_file(file_handle, 'rb')

        LOG.info("Loading taxonomy from file %s", file_handle.name)

        if '.msgpack' in file_handle.name:
            import msgpack
            merged = []
            for taxon_id, taxon in msgpack.Unpacker(file_handle, use_list=False, raw=False):
                taxon = TaxonTuple(*taxon)
                # if it's a merged taxon_id keep it and don't add it
                if taxon_id != taxon.taxon_id:
                    merged.append((taxon_id, taxon.taxon_id))
                else:
                    self[taxon_id] = taxon
            # the list of merged ids is iterated over and pointers are added to
            # the correct ids
            for merged_id, taxon_id in merged:
                self[merged_id] = self[taxon_id]

        else:
            self._taxa = pickle.load(file_handle)

    def save_data(self, file_handle):
        """
        .. versionchanged:: 0.2.3
            now can use *msgpack* to serialise

        Saves taxonomy data to a file handle or file name, can write compressed
        data if the file ends with ".gz", ".bz2"

        if the *.msgpack* string is found in the file name, the *msgpack*
        package is used instead of pickle

        Arguments:
            file_handle (str, file): file name to which save the instance data
        """

        if isinstance(file_handle, str):
            file_handle = open_file(file_handle, 'wb')

        LOG.info("Saving taxonomy to file %s", file_handle.name)

        if '.msgpack' in file_handle.name:
            import msgpack
            for taxon_id, taxon in viewitems(self._taxa):
                file_handle.write(
                    msgpack.packb((taxon_id, taxon))
                )
        else:
            pickle.dump(self._taxa, file_handle, -1)

    def find_by_name(self, s_name, rank=None, strict=True):
        """
        .. versionchanged:: 0.2.3
            the search is now case insensitive

        .. versionchanged:: 0.3.1
            added *rank* and *strict* parameter

        Returns the taxon IDs associated with the scientific name provided

        Arguments:
            s_name (str): the scientific name
            rank (str, None): return only a taxon_id of a specific rank
            strict (book): if True and *rank* is not None, KeyError will be
                raised if multiple taxa have the same name and rank

        Returns:
            list: a reference to the list of IDs that have for *s_name*, if
            *rank* is None. If *rank* is not None and one taxon is found, its
            taxon_id is returned, or None if no taxon is found. If *strict* is
            True and *rank* is not None, the set of taxon_ids found is
            resturned.

        Raises:
            KeyError: If multiple taxa are found, a *KeyError* exception is
            raised.
        """
        if not self._name_map:
            self.gen_name_map()

        if rank is None:
            return list(set(self._name_map[s_name.lower()]))
        else:
            taxon_ids = [
                taxon_id
                for taxon_id in self._name_map.get(s_name.lower(), [])
                if self[taxon_id].rank == rank
            ]
            if len(set(taxon_ids)) > 1:
                if strict:
                    raise KeyError(
                        "More than one taxon ({}) with rank {}".format(
                            s_name,
                            rank
                        )
                    )
                else:
                    return set(taxon_ids)
            elif len(taxon_ids) == 0:
                return None
            else:
                return taxon_ids[0]

    def is_ancestor(self, leaf_id, anc_ids):
        """
        .. versionchanged:: 0.1.13
            now uses :func:`is_ancestor` and changed behavior

        Checks if a taxon is the leaf of another one, or a list of taxa.

        :param int leaf_id: leaf taxon id
        :param int anc_ids: ancestor taxon id(s)

        :return bool: True if the ancestor taxon is in the leaf taxon lineage

        """

        if isinstance(anc_ids, int):
            anc_ids = [anc_ids]

        for anc_id in anc_ids:
            if is_ancestor(self, leaf_id, anc_id):
                return True
        return False

    def get_ranked_taxon(self, taxon_id, rank=None, ranks=TAXON_RANKS, roots=False):
        """
        .. versionchanged:: 0.1.13
            added *roots* argument

        Traverse the branch of which the *taxon* argument is the leaf backward,
        to get the specific rank to which the *taxon* belongs to.

        .. warning::

            the *roots* options is kept for backward compatibility and should be
            be set to *False*

        :param taxon_id: id of the taxon or instance of :class:`UniprotTaxon`
        :param str rank: string that specify the rank, if None, the first valid
            rank will be searched. (i.e. the first with a value different from '')
        :param ranks: tuple of all taxonomy ranks, default to the default module
            value
        :param bool roots: if True, uses :data:`TAXON_ROOTS` to solve the root
            taxa
        :return: instance of :class:`TaxonTuple` for the rank found.
        """

        if isinstance(taxon_id, int):
            ranked = self[taxon_id]
        else:
            raise ValueError("Not a valid taxon_id: {0}".format(taxon_id))

        while True:
            if (rank is None) and (ranked.rank != ''):
                break
            elif (ranked.rank == rank):
                break
            elif (ranked.rank in ranks) and \
                    (ranks.index(ranked.rank) < ranks.index(rank)):
                break
            elif ranked.parent_id is None:
                break
            # kept only for backward compatibility.
            # needs a check to other scripts
            elif roots and (ranked.s_name in TAXON_ROOTS):
                break

            ranked = self[ranked.parent_id]

        return ranked

    def get_ranked_id(self, taxon_id, rank=None, it=False, include_higher=True):
        """
        .. versionadded:: 0.3.4

        Gets the ranked taxon of another one. Useful when it's better to get a
        taxon_id instead of an instance of :class:`TaxonTuple`. Internally, it
        relies on :meth:`Taxonomy.get_ranked_taxon`.

        Arguments:
            taxon_id (int): taxon_id
            rank (str or None): passed over
            it (bool): determines the return value. if True, a list is returned
            include_higher (bool): if True, any rank higher than the requested
                may be returned. If False and the rank cannot be returned, None
                is returned

        Returns:
            int or list: The type returned is based on the *it* paramenter. If
            *it* is True, the return value is a list with the *taxon_id* of the
            ranked taxon as the sole value. If False, the returned value is the
            *taxon_id*. *include_higher* determines if the return value should
            be *None* if the exact rank was not found and *include_higher* is
            False
        """
        taxon = self.get_ranked_taxon(taxon_id, rank)
        if (taxon.rank != rank) and (not include_higher):
            taxon = None
        else:
            taxon = taxon.taxon_id
        if it:
            taxon = [taxon]
        return taxon

    def is_ranked_below(self, taxon_id, rank, equal=True):
        """
        .. versionadded:: 0.4.0

        Tests if a taxon_id is below the requested rank.

        Arguments:
            taxon_id (int): taxo_id to test
            rank (str): rank requested
            equal (bool): determines if the taxon_id tested may be of the
                requested rank

        Returns:
            bool: If the passed *taxon_id* is below the requested rank, it
            returns *True*. If *taxon_id* is of the rank requested and *equal*
            is True, the return value is *True*, if *equal* is False the return
            value is *False*. The return value is *False* otherwise.
        """
        if (self[taxon_id].rank == rank) and equal:
            return True

        taxon_id = self[taxon_id].parent_id

        while (taxon_id is not None):
            if self[taxon_id].rank == rank:
                return True
            else:
                taxon_id = self[taxon_id].parent_id

        return False

    def get_name_map(self):
        "Returns a taxon_id->s_name dictionary"

        return dict(
            (taxon.taxon_id, taxon.s_name)
            for taxon in self
        )

    def get_lineage(self, taxon_id, names=False, only_ranked=True, with_last=True):
        """
        .. versionadded:: 0.3.1

        Proxy for :func:`get_lineage`, with changed defaults

        Arguments:
            taxon_id (int): taxon_id to return the lineage
            names (bool): if the elements of the list are converted into the
                scientific names
            only_ranked (bool): only return the ranked taxa
            with_last (bool): include the taxon_id passed to the list

        Returns:
            list: the lineage of the passed taxon_id as a list of IDs or names
        """

        return get_lineage(
            self,
            taxon_id,
            names=names,
            only_ranked=only_ranked,
            with_last=with_last
        )

    def get_lineage_string(self, taxon_id, only_ranked=True, with_last=True,
                           sep=';', rank=None):
        """
        .. versionadded:: 0.3.3

        Generates a lineage string, with the possibility of getting another
        ranked taxon (via :meth:`Taxonomy.get_ranked_taxon`) to another
        rank, such as *phylum*.

        Arguments:
            taxon_id (int): taxon_id to return the lineage
            only_ranked (bool): only return the ranked taxa
            with_last (bool): include the taxon_id passed to the list
            sep (str): separator used to join the lineage string
            rank (int or None): if None the full lineage is returned, otherwise
                the lineage will be cut to the specified rank

        Returns:
            str: lineage string
        """

        if rank is not None:
            taxon_id = self.get_ranked_taxon(taxon_id, rank=rank).taxon_id

        return sep.join(
            self.get_lineage(
                taxon_id,
                only_ranked=only_ranked,
                with_last=with_last,
                names=True
            )
        )

    def add_taxon(self, taxon_name, common_name='', rank='no rank', parent_id=None):
        """
        .. versionadded:: 0.3.1

        Adds a taxon to the taxonomy. If a taxon with the same name and rank is
        found, its taxon_id is returned, otherwise a new taxon_id is returned.

        Arguments:
            taxon_name (str): scientific name of the taxon
            common_name (str): common name
            rank (str): rank, defaults to 'no rank'
            parent_id (int): taxon_id of the parent, defaults to *None*, which
                is the taxonomy root

        Returns:
            int: the taxon_id of the added taxon (if new), or the taxon_id of
            the taxon with the same name and rank found in the taxonomy

        Raises:
            KeyError: if more than one taxon has already the passed name and
            rank and it can't be resolved by looking at the parent_id passed,
            the exception is raised.
        """
        # Tries to find it in the name map
        taxon_id = self.find_by_name(taxon_name, rank=rank, strict=False)

        # Multiple taxa with the same name, check if the parent_id of any of
        # them can be resolved
        if isinstance(taxon_id, set):
            # print taxon_id
            taxon_ids = taxon_id
            taxon_id = None
            for t_id in taxon_ids:
                if self[t_id].parent_id == parent_id:
                    taxon_id = t_id

            if taxon_id is None:
                raise KeyError(
                    "Multiple taxa ({}, {}) are in the taxonomy but cannot be associated with the parent_id ({}) provided".format(
                        taxon_name,
                        rank,
                        parent_id
                    )
                )

        if (taxon_id is None) or (self[taxon_id].parent_id != parent_id):
            try:
                taxon_id = max(self._taxa) + 1
            except ValueError:
                taxon_id = 1

            self[taxon_id] = TaxonTuple(
                taxon_id,
                taxon_name,
                common_name,
                rank,
                (None,),
                parent_id
            )

            # adds the new name to the name map (which is initialised by
            # find_by_name)
            try:
                self._name_map[taxon_name.lower()].append(taxon_id)
            except KeyError:
                self._name_map[taxon_name.lower()] = [taxon_id]

        return taxon_id

    def add_lineage(self, **lineage):
        """
        .. versionadded:: 0.3.1

        Adds a lineage to the taxonomy. It's passed by keyword arguments, where
        each key is a value in the `TAXON_RANKS` rankes and the value is the
        scientific name. Appended underscores '_' will be stripped from the
        rank name. This is for cases such as *class* where the key is a reserved
        word in Python. Also one extra node can be added, such as
        strain/cultivar/subspecies and so on, but one only is expected to be
        passed.

        Arguments:
            lineage (dict): the lineage as a keyword arguments

        Returns:
            int: the taxon_id of the last element in the lineage

        Raises:
            ValueError: if more than a keyword argument is not contained in
            `TAXON_RANKS`
        """
        lineage = {
            key.rstrip('_'): value
            for key, value in viewitems(lineage)
        }
        extra_nodes = set(lineage) - set(TAXON_RANKS)
        if len(extra_nodes) > 1:
            raise ValueError(
                "Expected at most a one lower level below species, found: {}".format(
                    ', '.join(extra_nodes)
                )
            )
        else:
            extra_nodes = list(extra_nodes) if len(extra_nodes) == 1 else []

        ranks = list(TAXON_RANKS) + extra_nodes

        parent_ids = [None]
        taxon_id = None
        for rank in ranks:
            try:
                taxon_name = lineage[rank]
            except KeyError:
                continue

            if not isinstance(taxon_name, str):
                continue

            taxon_id = self.add_taxon(
                taxon_name,
                rank=rank,
                parent_id=parent_ids[-1]
            )
            # print taxon_id, rank, taxon_name, parent_ids[-1]
            parent_ids.append(taxon_id)

        return taxon_id

    def drop_taxon(self, taxon_id):
        """
        .. versionadded:: 0.3.1

        Drops a taxon and all taxa below it in the taxonomy. Also reset the
        name map for conistency.

        Arguments:
            taxon_id (int): taxon_id to drop from the taxonomy
        """
        drop_keys = []
        for leaf_id in self._taxa:
            if self.is_ancestor(leaf_id, taxon_id):
                drop_keys.append(leaf_id)

        for taxon_id in drop_keys:
            del self._taxa[taxon_id]

        self._name_map = {}

    def __getitem__(self, taxon_id):
        """
        Defines dictionary behavior. Key is a taxon_id, the returned value is a
        UniprotTaxon instance
        """
        return self._taxa[taxon_id]

    def __setitem__(self, taxon_id, taxon_obj):
        self._taxa[taxon_id] = taxon_obj

    def __delitem__(self, taxon_id):
        del self._taxa[taxon_id]

    def __iter__(self):
        """
        Defines iterable behavior. Returns a generator for UniprotTaxon instances
        """
        for taxon in viewvalues(self._taxa):
            yield taxon

    def __len__(self):
        """
        Returns the number of taxa contained
        """
        return len(self._taxa)

    def __contains__(self, taxon):
        """
        Returns True if the taxon is in the taxonomy

        Accepts an int (check for taxon_id) or an instance of UniprotTaxon
        """
        if isinstance(taxon, int):
            return taxon in self._taxa
        return False

    def __repr__(self):
        """
        .. versionadded:: 0.2.5
        """
        return "{} - {} taxa".format(self.__class__, len(self))


UniprotTaxonomy = Taxonomy


def get_ancestor_map(leaf_ids, anc_ids, taxonomy):
    """
    This function returns a dictionary where every leaf taxon is associated
    with the right ancestors in anc_ids

    ex. {clostridium: [bacteria, clostridia]}
    """
    anc_map = {}

    for leaf_id in leaf_ids:
        for anc_id in anc_ids:
            if taxonomy.is_ancestor(leaf_id, anc_id) or (leaf_id == anc_id):
                try:
                    anc_map[leaf_id].add(anc_id)
                except KeyError:
                    anc_map[leaf_id] = set()
                    anc_map[leaf_id].add(anc_id)

    return anc_map


def is_ancestor(taxonomy, taxon_id, anc_id):
    """
    .. versionchanged:: 0.1.16
        if a *taxon_id* raises a KeyError, False is returned

    Determine if the given taxon id (taxon_id) has anc_id as ancestor.

    :param :class:`Taxonomy` taxonomy: taxonomy used to test
    :param int taxon_id: leaf taxon to test
    :param int anc_id: ancestor taxon to test against

    :return bool: True if anc_id is an ancestor of taxon_id or their the same
    """
    if taxon_id == anc_id:
        return True

    while True:
        # if a taxon_id is not found, ancestry is invalid anyway, so it returns
        # False
        try:
            taxon_id = taxonomy[taxon_id].parent_id
        except KeyError:
            return False
        # print taxon_id, anc_id
        if anc_id == taxon_id:
            return True
        if taxon_id is None:
            return False


class NoLcaFound(Exception):
    """
    .. versionadded:: 0.1.13

    Raised if no lowest common ancestor can be found in the taxonomy
    """
    pass


def last_common_ancestor(taxonomy, taxon_id1, taxon_id2):
    """
    .. versionadded:: 0.1.13

    Finds the last common ancestor of two taxon IDs. An alias to this function
    is in the same module, called *lowest_common_ancestor* for compatibility.

    Arguments:
        taxonomy: :class:`Taxonomy` instance used to test
        taxon_id1 (int): first taxon ID
        taxon_id2 (int): second taxon ID

    Raturns:
        int: taxon ID of the lowest common ancestor

    Raises:
        NoLcaFound: if no common ancestor can be found
    """
    lca_id = taxon_id1

    while True:
        if is_ancestor(taxonomy, taxon_id2, lca_id):
            break
        else:
            lca_id = taxonomy[lca_id].parent_id
            if lca_id is None:
                raise NoLcaFound('No common ancestry')

    return lca_id


lowest_common_ancestor = last_common_ancestor


def distance_taxa_ancestor(taxonomy, taxon_id, anc_id):
    """
    .. versionadded:: 0.1.16

    Function to calculate the distance between a taxon and the given ancestor

    The distance is equal to the number of step in the taxonomy taken to arrive
    at the ancestor.

    Arguments:
        taxonomy: :class:`Taxonomy` instance
        taxon_id (int): taxonomic identifier
        anc_id (int): taxonomic identifier of the ancestor

    Raturns:
        int: distance between *taxon_id* and it ancestor *anc_id*

    """
    distance = 0

    while True:
        if anc_id == taxon_id:
            break
        taxon_id = taxonomy[taxon_id].parent_id
        distance += 1

    return distance


def distance_two_taxa(taxonomy, taxon_id1, taxon_id2):
    """
    .. versionadded:: 0.1.16

    Calculate the distance between two taxa. The distance is equal to the sum
    steps it takes to traverse the taxonomy until their last common ancestor.

    Arguments:
        taxonomy: :class:`Taxonomy` instance
        taxon_id1 (int): taxonomic identifier of first taxon
        taxon_id2 (int): taxonomic identifier of second taxon

    Raturns:
        int: distance between *taxon_id1* and *taxon_id2*
    """
    lca_id = last_common_ancestor(taxonomy, taxon_id1, taxon_id2)

    return sum(
        distance_taxa_ancestor(taxonomy, taxon_id, lca_id)
        for taxon_id in (taxon_id1, taxon_id2)
    )


def taxa_distance_matrix(taxonomy, taxon_ids):
    """
    .. versionadded:: 0.1.16

    Given a list of taxonomic identifiers, returns a distance matrix in a
    pairwise manner by using :func:`distance_two_taxa` on all possible
    two element combinations of *taxon_ids*.

    Arguments:
        taxonomy: :class:`Taxonomy` instance
        taxon_ids (iterable): list taxonomic identifiers

    Returns:
        pandas.DataFrame: matrix with the pairwise distances of all *taxon_ids*
    """
    import pandas

    matrix = pandas.DataFrame(index=taxon_ids, columns=taxon_ids).fillna(0)

    for taxon_id1, taxon_id2 in itertools.combinations(taxon_ids, 2):
        distance = distance_two_taxa(taxonomy, taxon_id1, taxon_id2)
        matrix.loc[taxon_id1, taxon_id2] = distance
        matrix.loc[taxon_id2, taxon_id1] = distance

    return matrix


def get_lineage(taxonomy, taxon_id, names=False, only_ranked=False,
                with_last=False):
    """
    .. versionadded:: 0.2.1

    .. versionchanged:: 0.2.5
        added *only_ranked*

    .. versionchanged:: 0.3.0
        added *with_last*

    Returns the lineage of a taxon_id, as a list of taxon_id or taxa names

    Arguments:
        taxonomy: a :class:`Taxonomy` instance
        taxon_id (int): taxon_id whose lineage to return
        names (bool): if True, the returned list contains the names of the taxa
            instead of the taxon_id
        only_ranked (bool): if True, only taxonomic levels whose rank is in
            data:`TAXON_RANKS` will be returned
        with_last (bool): if True, the passed taxon_id is included in the
            lineage

    Returns:
        list: lineage of the taxon_id, the elements are `int` if names is False,
        and `str` when *names* is True. If a taxon has no scientific name, the
        common name is used. If *only_ranked* is True, the returned list only
        contains ranked taxa (according to :data:`TAXON_RANKS`).
    """
    lineage = []

    if with_last:
        lineage.append(taxon_id)

    while True:
        taxon_id = taxonomy[taxon_id].parent_id
        if taxon_id is None:
            break
        if only_ranked and (taxonomy[taxon_id].rank not in TAXON_RANKS):
            continue
        lineage.append(taxon_id)

    if names:
        lineage = [
            taxonomy[x].s_name if taxonomy[x].s_name else taxonomy[x].c_name
            for x in lineage
        ]

    return list(reversed(lineage))


def last_common_ancestor_multiple(taxonomy, taxon_ids):
    """
    .. versionadded:: 0.2.5

    Applies :func:`last_common_ancestor` to an iterable that yields *taxon_id*
    while removing any *None* values. If the list is of one element, that
    *taxon_id* is returned.

    Arguments:
        taxonomy: instance of :class:`Taxonomy`
        taxon_ids (iterable): an iterable that yields taxon_id

    Returns:
        int: the *taxon_id* that is the last common ancestor of all taxon_ids
        passed

    Raises:
        NoLcaFound: when no common ancestry is found or the number of
        *taxon_ids* is 0
    """

    taxon_ids = [
        taxon_id
        for taxon_id in taxon_ids
        if taxon_id is not None
    ]

    if not taxon_ids:
        raise NoLcaFound('Empty *taxon_ids*')

    return reduce(
        lambda x1, x2: last_common_ancestor(taxonomy, x1, x2),
        set(taxon_ids)
    )



"""
SNPs data package
"""

# these import are for "legacy" scripts
# from .funcs import combine_snps_in_dataframe_test
# combine_snps_in_dataframe = combine_snps_in_dataframe_test

# imported for convenience
from .conv_func import get_rank_dataframe, get_gene_map_dataframe
from .funcs import combine_sample_snps

from . import filter
from . import funcs
from . import conv_func



"""
Various utilities to help read and process files
"""
from builtins import range, next
import logging
import gzip
import bz2
import io
import lzma

LOG = logging.getLogger(__name__)


def group_tuples_by_key(iterator, key_func=None, skip_elements=0):
    """
    .. versionadded:: 0.3.1

    Group the elements of an iterator by a key and yields the grouped elements.
    The elements yielded by the iterator are assumed to be a list or tuple,
    with the default key (when *key_func* is None) being the first of the of
    the objects inside that element. This behaviour can be customised by
    passing to *key_func* a function that accept an element and returns the key
    to be used.

    .. note::

        the iterable assumen that the elements are already sorted by their keys

    Arguments:
        iterator (iterable): iterator to be grouped
        key_func (func): function that accepts a element and returns its
            associated key
        skip_elements (int): number of elements to skip at the start

    Yields:
        list: a list of the grouped elements by key
    """
    if key_func is None:
        def key_func(x): return x[0]

    for index in range(skip_elements):
        next(iterator)

    curr_key = None
    curr_ann = []

    for element in iterator:
        new_key = key_func(element)
        if curr_key == new_key:
            curr_ann.append(element)
        else:
            if curr_key is None:
                curr_ann.append(element)
                curr_key = new_key
            else:
                yield curr_ann
                curr_key = new_key
                curr_ann = [element]
    else:
        yield curr_ann


class UnsupportedFormat(IOError):
    "Raised if the a file can't be opened with the correct module"
    pass


def open_file(file_name, mode='r'):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.3.4
        using *io.open*, always in binary mode

    .. versionchanged:: 0.4.2
        when a file handle is detected, it is passed to :func:`compressed_handle`
        to detect if the handle is a compressed file

    Opens a file using the extension as a guide to which module to use.

    .. note::

        Unicode makes for a slower `.translate` method in Python2, so it's
        best to use the `open` builtin.

    Arguments:
        file_name (str): file name
        mode (str): mode used to open the file

    Returns:
        file: file handle

    Raises:
        UnsupportedFormat: if the module to open the file is not available

    """

    test_class = io.IOBase

    if isinstance(file_name, test_class):
        return compressed_handle(file_name)

    mode = mode + 'b' if 'b' not in mode else mode

    if file_name.endswith('.gz'):
        file_handle = gzip.GzipFile(file_name, mode)
    elif file_name.endswith('.bz2'):
        file_handle = bz2.BZ2File(file_name, mode)
    elif file_name.endswith('.xz'):
        if lzma is None:
            raise UnsupportedFormat("Cannot import lzma module")
        else:
            file_handle = lzma.LZMAFile(file_name, mode)
    else:
        file_handle = io.open(file_name, mode)

    return file_handle


def compressed_handle(file_handle):
    """
    .. versionadded:: 0.1.13

    Tries to wrap a file handle in the appropriate compressed file class.

    Arguments:
        file_handle (str): file handle

    Returns:
        file: the same file handle if no suitable compressed file class is
        found or the new file_handle which supports the compression

    Raises:
        UnsupportedFormat: if the module to open the file is not available

    """
    # LOG.info('Detecting Compressed File Handle')

    file_name = getattr(file_handle, 'name', None)

    if file_name is None:
        return file_handle

    if file_name.endswith('.gz'):
        file_handle = gzip.GzipFile(fileobj=file_handle, mode='rb')
    elif file_name.endswith('.xz'):
        if lzma:
            raise UnsupportedFormat("Cannot import lzma module")
        else:
            file_handle = lzma.LZMAFile(file_handle)

    return file_handle


def split_write(records, name_mask, write_func, num_files=2):
    """
    .. versionadded:: 0.1.13

    Splits the writing of a number of records in a series of files. The
    `name_mask` is used as template for the file names. A string like
    "split-files-{0}" can be specified and the function applies format with the
    index of the pieces.

    Arguments:
        records (iterable): an iterable that returns a object to be saved
        name_mask (str): a string used as template for the output file names
            on which the function applies :func:`string.format`
        write_func (func): a function that is called to write to the files. It
            needs to accept a file handles as first argument and the record
            returned by `records` as the second argument
        num_files (int): the number of files to split the records
    """
    out_handles = [open_file(name_mask.format(x), 'w') for x in range(num_files)]

    for index, record in enumerate(records):
        out_handle = out_handles[index % num_files]
        write_func(out_handle, record)



"""
Module used to filter lists
"""

import numpy

from ..consts import MIN_COV


def aggr_filtered_list(val_list, aggr_func=numpy.mean,
                       filt_func=lambda x: False if x < MIN_COV else True):
    """
    Aggregate a list of values using 'aggr_func' on a list that passed the
    filtering in 'filt_func'.

    'filt_func' is a function that returns True or False for each value in
    val_list. If the return value is True, the element is included in the
    values passed to 'aggr_func'. Internally a list comprehension is used and
    the result passed to 'aggr_func'

    :param iterable val_list: list of values
    :param func aggr_func: function used to aggregate the list values
    :param func filt_func: function the return True or False

    :return: the result of the applied 'aggr_func'
    """
    return aggr_func([val for val in val_list if filt_func(val)])



"""
.. versionadded:: 0.2.3

This module defines routine to access Pfam information using a
network connection

"""
from . import url_open

PFAM_URL = "http://pfam.xfam.org/"


def get_pfam_families(key='id'):
    """
    .. versionadded:: 0.2.3

    Gets a dictionary with the accession/id/description of Pfam families
    from Pfam. This list can be accessed using the URL:
    http://pfam.xfam.org/families?output=text

    The output is a tab separated file where the fields are:

    * ACCESSION
    * ID
    * DESCRIPTION

    Arguments:
        key (str): if the value is *id*, the key of the dictionary is the ID,
            otherwise ID swaps position with ACCESSION (the new key)

    Returns:
        dict: by default the function returns a dictionary that uses the ID
        as key, while the value is a tuple (ACCESSION, DESCRIPTION). ID is the
        default because the :ref:`hmmer2gff` script output uses ID as *gene_id*
        value when using the HMM provided by Pfam
    """
    families = {}
    for line in url_open(PFAM_URL + "families?output=text", stream=True):
        line = line.decode('utf8').strip()
        if line.startswith('#') or (not line):
            continue
        acc, p_id, description = line.strip().split('\t')

        if key != 'id':
            p_id, acc = acc, p_id

        families[p_id] = (acc, description)

    return families



"Access EMBL Services"

from builtins import range
import logging
import re
import gzip
from . import url_read
import requests
import mgkit
try:
    from cStringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO


class NoEntryFound(Exception):
    """
    Raised if no sequences where found by :func:`get_sequences_by_ids`, the
    check is based on the :data:`NONE_FOUND` variable.
    """
    pass


class EntryNotFound(Exception):
    """
    Raised if at least one entry was not found by :func:`get_sequences_by_ids`.
    :data:`NOT_FOUND` is used to check if any entry wasn't downloaded.
    """
    pass


URL_REST = "http://www.ebi.ac.uk/Tools/dbfetch/dbfetch/"
"Base URL for EMBL DBFetch REST API"

URL_DATAWAREHOUSE = "http://www.ebi.ac.uk/ena/data/warehouse/search?"

EMBL_DBID = 'embl_cds'
"Default database id"

NOT_FOUND = r"Entry: .+? not found.\n"
"""
Appears in the resulting fasta (not tried on other formats) in the case that at
least one entry wasn't found.
"""

SUPPRESSED = "suppressed at the submitter's request on"

NONE_FOUND = r"ERROR 12.+?.\n?"
"""
Regular expression to check if no entry was found, used by :exc:`NoEntryFound`
"""

LOG = logging.getLogger(__name__)
"Log instance for this module"


def get_sequences_by_ids(embl_ids, contact=None, out_format='fasta',
                         num_req=10, embl_db=EMBL_DBID, strict=False):
    """
    .. versionchanged:: 0.3.4
        removed *compress* as it's bases on the `requests` package

    Downloads entries using EBI REST API. It can download one entry at a
    time or accept an iterable and all sequences will be downloaded in batches
    of at most num_req.

    It's fairly general, so can be customised, from the DB used to the output
    format: all batches are simply concatenate.

    .. note::

        There are some checks on the some errors reported by the EMBL api, but
        not documented, in particular two errors, which are just reported as
        text lines in the fasta file (the only one tested at this time).

        The are two possible cases:

        * if no entry was found :exc:`NoEntryFound` will be raised.
        * if at least one entry wasn't found:

          * if strict is False (the default) the error will be just logged as a
            debug message
          * if strict is True :exc:`EntryNotFound` is raised

    Args:
        embl_ids (iterable, str): list of ids to download
        contact (str): email address to be passed in the query
        format (str): format of the entry
        num_req (int): number of entries to download with each request
        embl_db (str): db to which the ids refer to
        strict (bool): if True, a check on the number of entries retrieved is
            performed

    Returns:
        str: the entries requested

    Raises:
        EntryNotFound: if at least an entry was not found
        NoEntryFound: if NO entry were found

    .. warning::

        The number of sequences that can be downloaded at a time is 11, it
        seems, since the returned sequences for each request was at most 11. I
        didn't find any mention of this in the API docs, but it may be a
        restriction that's temporary.

    """
    if isinstance(embl_ids, str):
        embl_ids = [embl_ids]
    elif isinstance(embl_ids, (set, dict)):
        embl_ids = list(embl_ids)

    splitter = re.compile(NOT_FOUND)
    error_re = re.compile(NONE_FOUND)

    entries = []

    for idx in range(0, len(embl_ids), num_req):

        if len(embl_ids) > num_req:
            LOG.debug("Downloading ids - range %d-%d", idx + 1, idx + num_req)

        request_params = "{db_id}/{entry_id}/{format}".format(
            db_id=embl_db,
            entry_id=','.join(embl_ids[idx:idx+num_req]),
            format=out_format
        )

        request_url = URL_REST + request_params

        seqs = url_read(request_url, agent=contact)

        # it seems that they add at the end of each request two newlines
        # (should be just one)
        seqs = seqs.replace('\n\n', '\n')

        if splitter.search(seqs):
            if strict:
                raise EntryNotFound("At least one Entry was not found")
            LOG.debug("At least one Entry was not found")
            entries.extend(splitter.split(seqs))
        else:
            entries.append(seqs)

    entries = ''.join(entries)

    # check for suppressed entries
    entries = '\n'.join(
        line for line in entries.split('\n')
        if line.find(SUPPRESSED) == -1
    )

    if error_re.search(entries):
        # if there's no sequences or we want all sequences to be downloaded
        if entries.find('>') == -1:
            raise NoEntryFound("No entry found")
        elif strict:
            raise EntryNotFound("At least one Entry was not found")
        # in case there's at least one sequence that was retrieved
        else:
            entries = ''.join(error_re.split(entries))
            LOG.debug("At least one Entry was not found")

    return entries


def dbfetch(embl_ids, db='embl', contact=None, out_format='seqxml',
            num_req=10):
    """
    .. versionadded:: 0.1.12

    Function that allows to use dbfetch service (REST). More information on the
    output formats and the database available at the
    `service page <http://www.ebi.ac.uk/Tools/dbfetch/syntax.jsp>`_

    Arguments:
        embl_ids (str, iterable): list or single sequence id to retrieve
        db (str): database from which retrieve the sequence data
        contact (str): email contact to use as per EMBL guidlines
        out_format (str): output format, depends on database
        num_req (int): number of ids per request

    Returns:
        list: a list with the results from each request sent. Each request sent
        has a maximum number *num_req* of ids, so the number of items in the
        list depends by the number of ids in *embl_ids* and the value of
        *num_req*.
    """
    if isinstance(embl_ids, str):
        embl_ids = [embl_ids]
    elif isinstance(embl_ids, (set, dict)):
        embl_ids = list(embl_ids)

    entries = []

    for idx in range(0, len(embl_ids), num_req):

        if len(embl_ids) > num_req:
            LOG.info("Downloading ids - range %d-%d", idx + 1, idx + num_req)

        request_params = "{db_id}/{entry_id}/{format}?style=raw".format(
            db_id=db,
            entry_id=','.join(embl_ids[idx:idx+num_req]),
            format=out_format
        )

        request_url = URL_REST + request_params

        if mgkit.DEBUG:
            LOG.debug(request_url)

        entries.append(url_read(request_url, agent=contact))

    return entries


def datawarehouse_search(query, domain='sequence', result='sequence_release',
                         display='fasta', offset=0, length=100000,
                         contact=None, download='gzip', url=URL_DATAWAREHOUSE,
                         fields=None):
    """
    .. versionchanged:: 0.2.3
        added *fields* parameter to retrieve tab separated information

    .. versionadded:: 0.1.13

    Perform a datawarehouse search on EMBL dbs. Instructions on the query
    language used to query the datawarehouse are available at `this page
    <http://www.ebi.ac.uk/ena/about/browser#data_warehouse>`_ with more details
    about the databases domains `at this page
    <http://www.ebi.ac.uk/ena/data/warehouse/usage>`_

    Arguments:
        query (str): query for the search enging
        domain (str): database domain to search
        result (str): domain result requested
        display (str): display option (format to retrieve the entries)
        offset (int): the offset of the search results, defaults to the first
        length (int): number of results to retrieve at the specified offset
            and the limit is automatically set a 100,000 records for query
        contact (str): email of the user
        download (str): type of response. Gzip responses are automatically
            decompressed
        url (str): base URL for the resource
        fields (None, iterable): must be an iterable of fields to be returned
            if display is set to *report*

    Returns:
        str: the raw request

    Examples:
        Querying EMBL for all sequences of type rRNA of the  *Clostridium*
        genus. Only from the EMBL release database in fasta format:

        >>> query = 'tax_tree(1485) AND mol_type="rRNA"'
        >>> result = 'sequence_release'
        >>> display = 'fasta'
        >>> data = embl.datawarehouse_search(query, result=result,
        ... display=display)
        >>> len(data)
        35919

        Each entry taxon_id from the same data can be retrieved using *report*
        as the *display* option and *fields* an iterable of fields to just
        ('accession', tax_id'):

        >>> query = 'tax_tree(1485) AND mol_type="rRNA"'
        >>> result = 'sequence_release'
        >>> display = 'report'
        >>> fields = ('accession', 'tax_id')
        >>> data = embl.datawarehouse_search(query, result=result,
            display=display, fields=fields)

    """

    params = {
        'query': query,
        'domain': domain,
        'result': result,
        'display': display,
        'offset': offset,
        'length': length,
        'download': download
    }

    # if display='report', fields can be used to display, but download needs to
    # be 'txt'
    if display == 'report':
        params['download'] = download = 'txt'
        try:
            params['fields'] = ','.join(fields)
        except TypeError:
            raise ValueError(
                "the 'fields' parameter must be an iteratble if display is "
                "equal to 'report'"
                )

    stream = True if download == 'gzip' else False
    request = requests.get(url, params, headers={'user-agent': contact}, stream=stream)
    if download == 'gzip':
        data = StringIO(request.raw.read())
        data = gzip.GzipFile(mode='rb', fileobj=data).read()
    else:
        data = request.text

    return data



"""
.. versionadded:: 0.1.14
"""

from . import colors
from . import heatmap
from . import boxplot
from . import utils

from .colors import *
from .heatmap import *
from .boxplot import *
from .utils import *
